"""
Some random notes on how Robot Framework does the running:

The main idea is that you create a TestSuite and run it.

Unfortunately, this isn't ideal for our use-case as the namespace is lost
during the running.

i.e.: in robot.running.suiterunner.SuiteRunner, in the `start_suite` method, a
Namespace is created and at that same place the imports for the TestSuite are
handled.

Afterwards, for each test it visits, it runs the needed setup/test/teardown loop.

But alas, we want to interactively add new library imports, resource imports,
keywords, so, the whole structure falls apart for the interpreter because the
visiting isn't really done per-statement, rather, the ast is collected and then
a bunch of internal structures are created and then the running is based on
those internal structures, not on the AST (probably a side-effect of having the
ast being added really late into Robot Framework and not from the start).

So, the approach being taken is the following:

1. Pre-create a test suite which will call a keyword where we'll pause to 
   actually execute the main loop.
   
2. In the main loop, collect the AST and then use the related builders to create
   the structure required to actually run, but instead of just blindly running it,
   verify what was actually loaded and dispatch accordingly (so, for instance,
   an import will use internal robot APIs to do the import -- and hopefully
   in the future when a usage is established, public APIs can be created in
   Robot Framework itself for this usage).

Previous work:

There is already a project which provides an interpreter:
    https://github.com/jupyter-xeus/robotframework-interpreter/blob/master/robotframework_interpreter/interpreter.py

    The approach used is that execute is done in blocks, so, it'll accept a full
    section and then execute it, copying back and forth the imports/variables/keywords
    between section evaluations.
"""
from robotframework_interactive.callbacks import Callback, CallbackWithReturn
import traceback
from ast import NodeVisitor
from robotframework_interactive.robotfacade import RobotFrameworkFacade
import sys
import os
from robotframework_interactive.protocols import (
    IOnReadyCall,
    EvaluateTextTypedDict,
    ActionResultDict,
)
import weakref

__file__ = os.path.abspath(__file__)
if __file__.endswith((".pyc", ".pyo")):
    __file__ = __file__[:-1]


class IOnOutput(object):
    def __call__(self, s: str):
        pass


class _CustomStream(object):
    def __init__(self, on_output: Callback):
        self._on_output = on_output

    def write(self, s):
        self._on_output(s)

    def flush(self):
        pass

    def isatty(self):
        return False


class _CustomStdIn:
    def __init__(self, interpreter, original_stdin=sys.stdin):
        self._interpreter = weakref.ref(interpreter)
        try:
            self.encoding = sys.stdin.encoding
        except:
            pass
        self.original_stdin = original_stdin

        try:
            self.errors = (
                sys.stdin.errors
            )  # Who knew? sys streams have an errors attribute!
        except:
            # Not sure if it's available in all Python versions...
            pass

    def readline(self, *args, **kwargs):
        interpreter = self._interpreter()
        if interpreter is not None:
            interpreter.on_before_read()

        try:
            return self.original_stdin.readline(*args, **kwargs)
        except KeyboardInterrupt:
            raise  # Let KeyboardInterrupt go through
        except:
            return "\n"
        finally:
            if interpreter is not None:
                interpreter.on_after_read()

    def write(self, *args, **kwargs):
        pass  # not available _CustomStdIn (but it can be expected to be in the stream interface)

    def flush(self, *args, **kwargs):
        pass  # not available _CustomStdIn (but it can be expected to be in the stream interface)

    def read(self, *args, **kwargs):
        # in the interactive interpreter, a read and a readline are the same.
        return self.readline()

    def close(self, *args, **kwargs):
        pass  # expected in _CustomStdIn

    def __iter__(self):
        # _CustomStdIn would not be considered as Iterable in Python 3 without explicit `__iter__` implementation
        return self.original_stdin.__iter__()

    def __getattr__(self, item):
        # it's called if the attribute wasn't found
        if hasattr(self.original_stdin, item):
            return getattr(self.original_stdin, item)
        raise AttributeError("%s has no attribute %s" % (self.original_stdin, item))


class _CustomErrorReporter(NodeVisitor):
    def __init__(self, source):
        self.source = source

    def visit_Error(self, node):
        facade = RobotFrameworkFacade()
        Token = facade.Token
        DataError = facade.DataError

        # All errors raise here!
        fatal = node.get_token(Token.FATAL_ERROR)
        if fatal:
            raise DataError(self._format_message(fatal))
        for error in node.get_tokens(Token.ERROR):
            raise DataError(self._format_message(error))

    def _format_message(self, token):
        return "Error in file '%s' on line %s: %s" % (
            self.source,
            token.lineno,
            token.error,
        )


class RobotFrameworkInterpreter(object):
    def __init__(self):
        from robotframework_interactive import main_loop

        main_loop.MainLoopCallbackHolder.ON_MAIN_LOOP = self.interpreter_main_loop
        facade = RobotFrameworkFacade()
        TestSuite = facade.TestSuite

        self._test_suite = TestSuite.from_file_system(
            os.path.join(os.path.dirname(__file__), "robot_interactive_console.robot")
        )
        self.on_stdout = Callback()
        self.on_stderr = Callback()
        self.on_exception_handled = CallbackWithReturn()
        self._stdout = _CustomStream(self.on_stdout)
        self._stderr = _CustomStream(self.on_stderr)

        self._waiting_for_input = 0
        weak_self = weakref.ref(self)

        def before_read():
            s = weak_self()
            if s is not None:
                s._waiting_for_input += 1

        def after_read():
            s = weak_self()
            if s is not None:
                s._waiting_for_input -= 1

        self.on_before_read = Callback()
        self.on_before_read.register(before_read)

        self.on_after_read = Callback()
        self.on_after_read.register(after_read)

        self._on_main_loop = None

        self._settings_section_name_to_block_mode = {
            "SettingSection": ("*** Settings ***\n", ""),
            "VariableSection": ("*** Variables ***\n", ""),
            "TestCaseSection": ("*** Test Case ***\nDefault Task/Test\n", "    "),
            "KeywordSection": ("*** Keyword ***\n", ""),
            "CommentSection": ("*** Comment ***\n", ""),
        }
        self._last_section_name = "TestCaseSection"
        self._last_block_mode_and_indent = self._settings_section_name_to_block_mode[
            self._last_section_name
        ]
        # the section we're tracking -> section ast
        self._doc_parts = {
            "CommentSection": None,
            "SettingSection": None,
            "VariableSection": None,
            "TestCaseSection": None,
            "KeywordSection": None,
        }

        sys.stdin = _CustomStdIn(self)

    @property
    def waiting_for_input(self):
        return bool(self._waiting_for_input)

    @property
    def full_doc(self) -> str:
        """
        :return:
            The full document as seen by the interpreter from what it was
            able to evaluate so far.

            Note that it should be logically consistent but not necessarily
            equal to what the user entered as statements.
        """
        return self._compute_full_doc()

    def _compute_full_doc(self, last_section_name=""):
        from robotframework_interactive.ast_to_code import ast_to_code

        full_doc = []
        sections = [
            "CommentSection",
            "SettingSection",
            "VariableSection",
            "KeywordSection",
            "TestCaseSection",
        ]
        if last_section_name:
            sections.remove(last_section_name)
            sections.append(last_section_name)

        for part in sections:
            part_as_ast = self._doc_parts[part]
            if part_as_ast:
                as_code = ast_to_code(part_as_ast).strip()
                if as_code:
                    full_doc.append(as_code)
        return ("\n".join(full_doc)).strip()

    def initialize(self, on_main_loop: IOnReadyCall):
        self._on_main_loop = on_main_loop

        stdout = self._stdout
        stderr = self._stderr
        options = dict(
            output=os.path.abspath("output.xml"), stdout=stdout, stderr=stderr
        )

        self._test_suite.run(**options)

    def interpreter_main_loop(self, *args, **kwargs):
        self._on_main_loop(self)

    def compute_evaluate_text(
        self, code: str, target_type: str = "evaluate"
    ) -> EvaluateTextTypedDict:
        """
        :param target_type:
            'evaluate': means that the target is an evaluation with the given code.
                This implies that the current code must be changed to make sense
                in the given context.

            'completions': means that the target is a code-completion
                This implies that the current code must be changed to include
                all previous evaluation so that the code-completion contains
                the full information up to the current point.
        """
        ret: EvaluateTextTypedDict = {"prefix": "", "full_code": code, "indent": ""}

        if target_type == "completions":
            if code.strip().startswith("*"):
                # easy mode, just get the full doc and concatenate it with the code.
                prefix = self.full_doc + "\n"
                ret = {"prefix": prefix, "full_code": prefix + code, "indent": ""}
            else:
                # Ok, we need to see how the current code would fit in with the
                # existing code.
                last_section_name = self._last_section_name
                has_section = self._doc_parts.get(last_section_name) is not None
                if has_section:
                    prefix = self._compute_full_doc(last_section_name)
                    first_part, delimiter, last_line = prefix.rpartition("\n")
                    if delimiter:
                        whitespaces = []
                        for c in last_line:
                            if c in ("\t", " "):
                                whitespaces.append(c)
                            else:
                                break
                        indent = "".join(whitespaces)

                        if not indent:
                            # Corner case: the user entered:
                            # *** Tasks ***
                            # Task name
                            #
                            # but didn't add any keyword call -- we have to
                            # indent ourselves here.
                            _last_mode, indent = self._last_block_mode_and_indent
                            prefix = self.full_doc + "\n"
                            indented_code = indent + ("\n" + indent).join(
                                code.split("\n")
                            )
                            return {
                                "prefix": prefix,
                                "full_code": prefix + indented_code,
                                "indent": indent,
                            }

                        prefix = first_part + delimiter + last_line + "\n"
                        indented_code = indent + ("\n" + indent).join(code.split("\n"))
                        ret = {
                            "prefix": prefix,
                            "full_code": prefix + indented_code,
                            "indent": indent,
                        }

                    else:
                        ret = {
                            "prefix": prefix,
                            "full_code": prefix + "\n" + code,
                            "indent": indent,
                        }
                else:
                    # There's no entry for this kind of section so far, so, we
                    # need to get the full block mode.
                    last_mode, indent = self._last_block_mode_and_indent
                    prefix = self.full_doc + "\n" + last_mode
                    indented_code = indent + ("\n" + indent).join(code.split("\n"))
                    ret = {
                        "prefix": prefix,
                        "full_code": prefix + indented_code,
                        "indent": indent,
                    }

        else:
            if not code.strip().startswith("*"):
                last_mode, indent = self._last_block_mode_and_indent
                indented_code = indent.join(code.split("\n"))
                ret = {
                    "prefix": last_mode,
                    "indent": indent,
                    "full_code": last_mode + indent + indented_code,
                }

        return ret

    def evaluate(self, code: str) -> ActionResultDict:
        original_stdout = sys.__stdout__
        original_stderr = sys.__stderr__
        try:
            # When writing to the console, RF uses sys.__stdout__, so, we
            # need to hijack it too...
            sys.__stdout__ = self._stdout
            sys.__stderr__ = self._stderr
            return self._evaluate(code)
        except Exception as e:
            if not self.on_exception_handled(e):
                # If it's not handled by some client, print it to stderr.
                s = traceback.format_exc()
                if s:
                    for line in s.splitlines(keepends=True):
                        self.on_stderr(line)
            return {
                "success": False,
                "message": f"Error while evaluating: {e}",
                "result": None,
            }
        finally:
            sys.__stdout__ = original_stdout
            sys.__stderr__ = original_stderr

    def _evaluate(self, code: str) -> ActionResultDict:
        # Compile AST
        from io import StringIO
        from robot.api import Token

        facade = RobotFrameworkFacade()
        get_model = facade.get_model
        TestSuite = facade.TestSuite
        TestDefaults = facade.TestDefaults
        SettingsBuilder = facade.SettingsBuilder
        EXECUTION_CONTEXTS = facade.EXECUTION_CONTEXTS
        SuiteBuilder = facade.SuiteBuilder

        code = self.compute_evaluate_text(code)["full_code"]

        model = get_model(
            StringIO(code),
            data_only=False,
            curdir=os.path.abspath(os.getcwd()).replace("\\", "\\\\"),
        )

        if not model.sections:
            msg = "Unable to interpret: no sections found."
            self.on_stderr(msg)
            return {
                "success": False,
                "message": f"Error while evaluating: {msg}",
                "result": None,
            }

        # Raise an error if there's anything wrong in the model that was parsed.
        _CustomErrorReporter(code).visit(model)

        # Initially it was engineered so that typing *** Settings *** would enter
        # *** Settings *** mode, but this idea was abandoned (it's implementation
        # is still here as we may want to revisit it, but it has some issues
        # in how to compute the full doc for code-completion, so, the default
        # section is always a test-case section now).
        #
        # last_section = model.sections[-1]
        # last_section_name = last_section.__class__.__name__
        # last_section_name = "TestCaseSection"
        # block_mode = self._settings_section_name_to_block_mode.get(last_section_name)
        # if block_mode is None:
        #     self.on_stderr(f"Unable to find block mode for: {last_section_name}")
        #
        # else:
        #     self._last_block_mode_and_indent = block_mode
        #     self._last_section_name = last_section_name

        new_suite = TestSuite(name="Default test suite")
        defaults = TestDefaults()

        SettingsBuilder(new_suite, defaults).visit(model)
        SuiteBuilder(new_suite, defaults).visit(model)

        # ---------------------- handle what was loaded in the settings builder.
        current_context = EXECUTION_CONTEXTS.current
        namespace = current_context.namespace
        source = os.path.join(
            os.path.abspath(os.getcwd()), "in_memory_interpreter.robot"
        )
        for new_import in new_suite.resource.imports:
            new_import.source = source
            # Actually do the import (library, resource, variable)
            namespace._import(new_import)

        if new_suite.resource.variables:
            # Handle variables defined in the current test.
            for variable in new_suite.resource.variables:
                variable.source = source

            namespace.variables.set_from_variable_table(new_suite.resource.variables)

        if new_suite.resource.keywords:
            # It'd be really nice to have a better API for this...
            user_keywords = namespace._kw_store.user_keywords
            for kw in new_suite.resource.keywords:
                kw.actual_source = source
                handler = user_keywords._create_handler(kw)

                embedded = isinstance(handler, facade.EmbeddedArgumentsHandler)
                if not embedded:
                    if handler.name in user_keywords.handlers._normal:
                        del user_keywords.handlers._normal[handler.name]
                user_keywords.handlers.add(handler, embedded)

        # --------------------------------------- Actually run any test content.
        for test in new_suite.tests:
            context = EXECUTION_CONTEXTS.current
            facade.run_test_body(context, test)

        # Now, update our representation of the document to include what the
        # user just entered.
        for section in model.sections:
            section_name = section.__class__.__name__
            if section.body:
                if section_name not in self._doc_parts:
                    continue

                current = self._doc_parts[section_name]
                if not current:
                    add = True
                    if section.__class__.__name__ == "TestCaseSection" and (
                        not section.body
                        or (len(section.body) == 1 and not section.body[0].body)
                    ):
                        add = False
                    if add:
                        current = self._doc_parts[section_name] = section
                else:
                    if current.__class__.__name__ == "TestCaseSection":
                        current = current.body[-1]
                        for test_case in section.body:
                            current.body.extend(test_case.body)
                    else:
                        current.body.extend(section.body)

                if current is not None:
                    # Make sure that there is a '\n' as the last EOL.
                    last_in_body = current.body[-1]
                    while not hasattr(last_in_body, "tokens"):
                        last_in_body = last_in_body.body[-1]
                    tokens = last_in_body.tokens
                    last_token = tokens[-1]
                    found_new_line = False
                    if last_token.type == Token.EOL:
                        if not last_token.value:
                            last_token.value = "\n"
                            found_new_line = True
                    if not found_new_line:
                        last_in_body.tokens += (Token("EOL", "\n"),)

        return {"success": True, "message": None, "result": None}



def test_dump_threads():
    import pydevd
    from io import StringIO
    stream = StringIO()
    pydevd.dump_threads(stream=stream)
    contents = stream.getvalue()
    assert 'Thread MainThread  (daemon: False, pydevd thread: False)' in contents
    assert 'test_dump_threads' in contents



import sys
import threading
from typing import (
    Dict,
    Union,
    Any,
    Generic,
    Callable,
    Mapping,
    Optional,
    List,
    Type,
    Iterable,
    Tuple,
)
from typing import TypeVar
import typing

from enum import Enum


if typing.TYPE_CHECKING:
    # This would lead to a circular import, so, do it only when type-checking.
    from robocorp_ls_core.callbacks import Callback
    from robocorp_ls_core.lsp import TextDocumentContentChangeEvent
    from robocorp_ls_core.lsp import HoverResponseTypedDict
    from robocorp_ls_core.lsp import ReferencesResponseTypedDict
    from robocorp_ls_core.lsp import TextDocumentTypedDict
    from robocorp_ls_core.lsp import ResponseTypedDict
    from robocorp_ls_core.lsp import CodeLensTypedDict
    from robocorp_ls_core.lsp import RangeTypedDict
    from robocorp_ls_core.lsp import DocumentHighlightResponseTypedDict
    from robocorp_ls_core.lsp import PositionTypedDict
    from robocorp_ls_core.lsp import CompletionItemTypedDict
    from robocorp_ls_core.lsp import CompletionsResponseTypedDict
    from robocorp_ls_core.lsp import CompletionResolveResponseTypedDict
    from robocorp_ls_core.lsp import TextDocumentItem

# Hack so that we don't break the runtime on versions prior to Python 3.8.
if sys.version_info[:2] < (3, 8):

    class Protocol(object):
        pass

    class TypedDict(object):
        def __init_subclass__(self, *args, **kwargs):
            pass


else:
    from typing import Protocol
    from typing import TypedDict


T = TypeVar("T")
Y = TypeVar("Y", covariant=True)


class Sentinel(Enum):
    SENTINEL = 0
    USE_DEFAULT_TIMEOUT = 1


def check_implements(x: T) -> T:
    """
    Helper to check if a class implements some protocol.

    :important: It must be the last method in a class due to
                https://github.com/python/mypy/issues/9266

        Example:

    def __typecheckself__(self) -> None:
        _: IExpectedProtocol = check_implements(self)

    Mypy should complain if `self` is not implementing the IExpectedProtocol.
    """
    return x


class IFuture(Generic[Y], Protocol):
    def result(self, timeout: typing.Optional[int] = None) -> Y:
        """Return the result of the call that the future represents.

        Args:
            timeout: The number of seconds to wait for the result if the future
                isn't done. If None, then there is no limit on the wait time.

        Returns:
            The result of the call that the future represents.

        Raises:
            CancelledError: If the future was cancelled.
            TimeoutError: If the future didn't finish executing before the given
                timeout.
            Exception: If the call raised then that exception will be raised.
        """

    def add_done_callback(self, fn: Callable[["IFuture"], Any]):
        """Attaches a callable that will be called when the future finishes.

        Args:
            fn: A callable that will be called with this future as its only
                argument when the future completes or is cancelled. The callable
                will always be called by a thread in the same process in which
                it was added. If the future has already completed or been
                cancelled then the callable will be called immediately. These
                callables are called in the order that they were added.
        """


class IEndPoint(Protocol):
    def notify(self, method: str, params: Any = None):
        """Send a JSON RPC notification to the client.

        Args:
            method (str): The method name of the notification to send
            params (any): The payload of the notification
        """

    def request(self, method: str, params=None) -> IFuture:
        """Send a JSON RPC request to the client.

        Args:
            method (str): The method name of the message to send
            params (any): The payload of the message

        Returns:
            Future that will resolve once a response has been received
        """

    def consume(self, message: Dict):
        """Consume a JSON RPC message from the client.

        Args:
            message (dict): The JSON RPC message sent by the client
        """


class IProgressReporter(Protocol):
    def set_additional_info(self, additional_info: str) -> None:
        """
        The progress reporter shows the title and elapsed time automatically.

        With this API it's possible to add additional info for the user to see.
        """


class CommunicationDropped(object):
    def __str__(self):
        return "CommunicationDropped"

    __repr__ = __str__


class IMessageMatcher(Generic[T], Protocol):

    event: threading.Event
    msg: T


class IIdMessageMatcher(Generic[T], Protocol):

    message_id: str
    event: threading.Event
    msg: T


COMMUNICATION_DROPPED = CommunicationDropped()


class IRequestCancellable(Protocol):
    def request_cancel(self, message_id) -> None:
        """
        Requests that some processing is cancelled.
        """


class IRequestHandler(Protocol):
    def __call__(self, request_name: str, msg_id: Any, params: Any) -> bool:
        """
        :param request_name:
            The name of the request to be handled.

        :param msg_id:
            The id of the message (to which a response should be generated).

        :param params:
            The parameters received in the request.

        :return:
            True if the request was handled (in which case, if multiple request
            handlers are registered, others aren't processed anymore) and False
            otherwise.
        """


class ILanguageServerClientBase(IRequestCancellable, Protocol):

    on_message: "Callback"

    def request_async(self, contents: Dict) -> Optional[IIdMessageMatcher]:
        """
        API which allows to wait for the message to complete.

        To use:
            message_matcher = client.request_async(contents)
            if message_matcher is not None:
                if message_matcher.event.wait(5):
                    ...
                    msg = message_matcher.msg
                else:
                    # Timed out

        :param contents:
            Something as:
            {"jsonrpc": "2.0", "id": msg_id, "method": method_name, "params": params}

        :return _MessageMatcher:
        """

    def request(
        self,
        contents,
        timeout: Union[int, Sentinel] = Sentinel.USE_DEFAULT_TIMEOUT,
        default: Any = COMMUNICATION_DROPPED,
    ):
        """
        :param contents:
        :param timeout:
        :return:
            The returned message if everything goes ok.
            `default` if the communication dropped in the meanwhile and timeout was None.

        :raises:
            TimeoutError if the timeout was given and no answer was given at the available time
            (including if the communication was dropped).
        """

    def obtain_pattern_message_matcher(
        self, message_pattern: Dict[str, str]
    ) -> IMessageMatcher:
        """
        Can be used as:

        message_matcher = language_server.obtain_pattern_message_matcher(
            {"method": "textDocument/publishDiagnostics"}
        )
        """

    def obtain_id_message_matcher(self, message_id) -> IMessageMatcher:
        pass

    def register_request_handler(self, message: str, handler: IRequestHandler) -> None:
        pass

    def write(self, contents):
        pass

    def shutdown(self):
        pass

    def exit(self):
        pass


class IRobotFrameworkApiClient(ILanguageServerClientBase, Protocol):
    def initialize(
        self, msg_id=None, process_id=None, root_uri="", workspace_folders=()
    ):
        pass

    def get_version(self) -> str:
        pass

    def lint(self, doc_uri: str) -> "ResponseTypedDict":
        pass

    def request_lint(self, doc_uri: str) -> Optional[IIdMessageMatcher]:
        pass

    def request_semantic_tokens_full(
        self, text_document: "TextDocumentTypedDict"
    ) -> Optional[IIdMessageMatcher]:
        pass

    def forward(self, method_name, params):
        pass

    def forward_async(self, method_name, params) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def open(self, uri, version, source):
        pass

    def request_section_name_complete(
        self, doc_uri, line, col
    ) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def request_keyword_complete(
        self, doc_uri, line, col
    ) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def request_complete_all(
        self, doc_uri, line, col
    ) -> Optional[IIdMessageMatcher["CompletionsResponseTypedDict"]]:
        """
        Completes: sectionName, keyword, variables
        :Note: async complete.
        """

    def request_resolve_completion_item(
        self, completion_item: "CompletionItemTypedDict"
    ) -> Optional[IIdMessageMatcher["CompletionResolveResponseTypedDict"]]:
        pass

    def request_find_definition(
        self, doc_uri, line, col
    ) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def request_source_format(
        self, text_document, options
    ) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def request_signature_help(
        self, doc_uri: str, line: int, col: int
    ) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def request_folding_range(self, doc_uri: str) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def request_hover(
        self, doc_uri: str, line: int, col: int
    ) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def request_references(
        self, doc_uri: str, line: int, col: int, include_declaration: bool
    ) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def request_workspace_symbols(
        self, query: Optional[str] = None
    ) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def settings(self, settings: Dict):
        pass

    def request_wait_for_full_test_collection(self) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """

    def request_evaluatable_expression(
        self, doc_uri: str, position: "PositionTypedDict"
    ) -> Optional[IIdMessageMatcher]:
        """
        :Note: async complete.
        """


class ITestInfoTypedDict(TypedDict):
    uri: str
    path: str
    name: str
    range: "RangeTypedDict"


class ITestInfoFromSymbolsCacheTypedDict(TypedDict):
    name: str
    range: "RangeTypedDict"


class ITestInfoFromUriTypedDict(TypedDict):
    uri: str
    testInfo: List[ITestInfoFromSymbolsCacheTypedDict]


class ILanguageServerClient(ILanguageServerClientBase, Protocol):
    pid: Optional[int]

    DEFAULT_TIMEOUT: Optional[int] = None

    def settings(self, settings: Dict):
        """
        :param settings:
            Something as:
            {"settings": {"robot": {"pythonpath": [case4_path]}}}
        """

    def initialize(
        self, root_path: str, msg_id=None, process_id=None, initialization_options=None
    ):
        pass

    def change_workspace_folders(
        self, added_folders: List[str], removed_folders: List[str]
    ) -> None:
        pass

    def open_doc(self, uri: str, version: int = 1, text: Optional[str] = None):
        """
        :param text:
            If None (default), the contents will be loaded from the disk.
        """

    def close_doc(self, uri: str):
        pass

    def change_doc(self, uri: str, version: int, text: str):
        pass

    def get_completions(
        self, uri: str, line: int, col: int
    ) -> "CompletionsResponseTypedDict":
        """
        :param uri:
            The uri for the request.
        :param line:
            0-based line.
        :param col:
            0-based col.
        """

    def get_completions_async(
        self, uri: str, line: int, col: int
    ) -> Optional[IIdMessageMatcher["CompletionsResponseTypedDict"]]:
        """
        :param uri:
            The uri for the request.
        :param line:
            0-based line.
        :param col:
            0-based col.
        """

    def request_source_format(self, uri: str):
        """
        :param uri:
            The uri for the request.
        """

    def find_definitions(self, uri, line: int, col: int):
        """
        :param uri:
            The uri for the request.
        :param line:
            0-based line.
        :param col:
            0-based col.
        """

    def execute_command(self, command: str, arguments: list) -> Mapping[str, Any]:
        pass

    def execute_command_async(
        self, command: str, arguments: list
    ) -> Optional[IIdMessageMatcher]:
        pass

    def request_signature_help(self, uri: str, line: int, col: int):
        pass

    def request_hover(self, uri: str, line: int, col: int) -> "HoverResponseTypedDict":
        pass

    def request_text_document_highlight(
        self, uri: str, line: int, col: int
    ) -> "DocumentHighlightResponseTypedDict":
        pass

    def request_references(
        self, uri: str, line: int, col: int, include_declaration: bool
    ) -> "ReferencesResponseTypedDict":
        pass

    def request_folding_range(self, uri: str):
        pass

    def request_code_lens(self, uri: str):
        pass

    def request_resolve_code_lens(self, code_lens: "CodeLensTypedDict"):
        pass

    def request_document_symbol(self, uri: str):
        pass

    def request_workspace_symbols(self, query: Optional[str] = None):
        pass

    def request_resolve_completion(
        self, completion_item: "CompletionItemTypedDict"
    ) -> "CompletionResolveResponseTypedDict":
        pass

    def hover(self, uri: str, line: int, col: int):
        """
        :param uri:
            The uri for the request.
        :param line:
            0-based line.
        :param col:
            0-based col.
        """


class IConfig(Protocol):
    def get_setting(
        self, key: str, expected_type: Any, default=Sentinel.SENTINEL
    ) -> Any:
        """
        :param key:
            The setting to be gotten (i.e.: my.setting.to.get)

        :param expected_type:
            The type which we're expecting.

        :param default:
            If given, return this value instead of throwing a KeyError.

        :raises:
            KeyError if the setting could not be found and default was not provided.
        """

    def update(self, settings: dict) -> None:
        """Recursively merge the given settings into the current settings."""

    def get_full_settings(self) -> dict:
        pass

    def set_override_settings(self, override_settings: dict) -> None:
        """
        Used to override settings with the keys given (note: any existing
        override setting will be removed and all the keys here will be set to
        override the initial settings -- use update_override_settings to keep
        other existing overrides).v
        """

    def update_override_settings(self, override_settings: dict) -> None:
        """
        Used to update existing override settings with the keys given.
        """

    def set_workspace_dir(self, workspace: str) -> None:
        pass


class ILog(Protocol):
    def critical(self, msg: str = "", *args: Any):
        pass

    def info(self, msg: str = "", *args: Any):
        pass

    def warn(self, msg: str = "", *args: Any):
        pass  # same as info

    def warning(self, msg: str = "", *args: Any):
        pass  # same as info

    def debug(self, msg: str = "", *args: Any):
        pass

    def exception(self, msg: str = "", *args: Any):
        pass

    def error(self, msg: str = "", *args: Any):
        pass  # same as exception


class IConfigProvider(Protocol):
    @property
    def config(self) -> IConfig:
        pass


class ILanguageServer(IConfigProvider):
    pass


class IDirCache(Protocol):
    """ """

    def store(self, key: Any, value: Any) -> None:
        """
        Persists the given key and value.

        :param key:
            The key to be persisted. It's repr(key) is used to calculate
            the key filename on the disk.

        :note that the values do a round-trip with json (so, caveats
        such as saving a tuple and loading a list apply).
        """

    def load(self, key: Any, expected_class: Type) -> Any:
        """
        Loads a previously persisted value.

        If it doesn't exist, there's some error loading or the expected
        class doesn't match the loaded value a KeyError is thrown.

        :note: users should check that the cache value is what's expected when it's
           gotten (as the data may become corrupted on disk or may change across
           versions).
        """

    def discard(self, key: Any) -> None:
        """
        Removes some key from the cache.
        """


class IDocumentSelection(Protocol):

    doc: "IDocument"
    line: int
    col: int

    @property
    def offset_at_position(self):
        """Return the byte-offset pointed at by the given position."""
        pass

    @property
    def current_line(self) -> str:
        pass

    @property
    def line_to_column(self):
        pass

    @property
    def word_at_column(self) -> str:
        pass

    @property
    def word_to_column(self) -> str:
        pass

    @property
    def word_from_column(self) -> str:
        pass


class IDocument(Protocol):
    uri: str
    version: Optional[str]
    path: str

    immutable: bool

    def selection(self, line: int, col: int) -> IDocumentSelection:
        pass

    @property
    def source(self) -> str:
        pass

    @source.setter
    def source(self, source: str) -> None:
        pass

    def get_line(self, line: int) -> str:
        pass

    def offset_to_line_col(self, offset: int) -> Tuple[int, int]:
        pass

    def get_range(self, line: int, col: int, endline: int, endcol: int) -> str:
        pass

    def get_last_line(self) -> str:
        pass

    def get_last_line_col(self) -> Tuple[int, int]:
        pass

    def get_last_line_col_with_contents(self, contents: str) -> Tuple[int, int]:
        pass

    def get_line_count(self) -> int:
        pass

    def apply_change(self, change: "TextDocumentContentChangeEvent") -> None:
        pass

    def is_source_in_sync(self) -> bool:
        """
        If the document is backed up by a file, returns true if the sources are
        in sync.
        """

    def find_line_with_contents(self, contents: str) -> int:
        pass


class IWorkspaceFolder(Protocol):
    uri: str
    name: str


class IWorkspace(Protocol):
    on_file_changed: "Callback"

    def on_changed_config(self, config: IConfig) -> None:
        pass

    @property
    def root_path(self):
        pass

    @property
    def root_uri(self):
        pass

    def iter_documents(self) -> Iterable[IDocument]:
        """
        Note: can only be called in the main thread.
        """

    def iter_folders(self) -> Iterable[IWorkspaceFolder]:
        """
        Note: the lock must be obtained when iterating folders.
        """

    def remove_document(self, uri: str) -> None:
        pass

    def put_document(self, text_document: "TextDocumentItem") -> IDocument:
        pass

    def get_document(self, doc_uri: str, accept_from_file: bool) -> Optional[IDocument]:
        """
        Return a managed document if-present, otherwise, create one pointing at
        the disk if accept_from_file == True (if the file exists, and we're able to
        load it, otherwise, return None).
        """

    def get_folder_paths(self) -> List[str]:
        """
        Retuns the folders which are set as workspace folders.
        """

    def dispose(self):
        pass


class ITimeoutHandle(Protocol):
    def exec_on_timeout(self):
        pass


class IMonitor(Protocol):
    def cancel(self) -> None:
        pass

    def check_cancelled(self) -> None:
        """
        raises JsonRpcRequestCancelled if cancelled.
        """


class ActionResultDict(TypedDict):
    success: bool
    message: Optional[
        str
    ]  # if success == False, this can be some message to show to the user
    result: Any


class ActionResult(Generic[T]):

    success: bool
    message: Optional[
        str
    ]  # if success == False, this can be some message to show to the user
    result: Optional[T]

    def __init__(
        self, success: bool, message: Optional[str] = None, result: Optional[T] = None
    ):
        self.success = success
        self.message = message
        self.result = result

    def as_dict(self) -> ActionResultDict:
        return {"success": self.success, "message": self.message, "result": self.result}

    def __str__(self):
        return f"ActionResult(success={self.success!r}, message={self.message!r}, result={self.result!r})"

    __repr__ = __str__


class EnvEntry(TypedDict):
    key: str
    value: str


class LibraryVersionDict(TypedDict):
    library: str
    version: str


class LibraryVersionInfoDict(TypedDict):
    success: bool

    # if success == False, this can be some message to show to the user
    message: Optional[str]

    # Note that if the library was found but the version doesn't match, the
    # result should still be provided.
    result: Optional[LibraryVersionDict]



class MyDictSubclass(dict):

    def __init__(self):
        dict.__init__(self)
        self.var1 = 10
        self['in_dct'] = 20

    def __str__(self):
        ret = []
        for key, val in sorted(self.items()):
            ret.append('%s: %s' % (key, val))
        ret.append('self.var1: %s' % (self.var1,))
        return '{' + '; '.join(ret) + '}'

    __repr__ = __str__


class MyListSubclass(list):

    def __init__(self):
        list.__init__(self)
        self.var1 = 11
        self.append('a')
        self.append('b')

    def __str__(self):
        ret = []
        for obj in self:
            ret.append(repr(obj))
        ret.append('self.var1: %s' % (self.var1,))
        return '[' + ', '.join(ret) + ']'

    __repr__ = __str__


class MySetSubclass(set):

    def __init__(self):
        set.__init__(self)
        self.var1 = 12
        self.add('a')

    def __str__(self):
        ret = []
        for obj in sorted(self):
            ret.append(repr(obj))
        ret.append('self.var1: %s' % (self.var1,))
        return 'set([' + ', '.join(ret) + '])'

    __repr__ = __str__


class MyTupleSubclass(tuple):

    def __new__ (cls):
        return super(MyTupleSubclass, cls).__new__(cls, tuple(['a', 1]))

    def __init__(self):
        self.var1 = 13

    def __str__(self):
        ret = []
        for obj in self:
            ret.append(repr(obj))
        ret.append('self.var1: %s' % (self.var1,))
        return 'tuple(' + ', '.join(ret) + ')'

    __repr__ = __str__


def Call():
    variable_for_test_1 = MyListSubclass()
    variable_for_test_2 = MySetSubclass()
    variable_for_test_3 = MyDictSubclass()
    variable_for_test_4 = MyTupleSubclass()

    all_vars_set = True  # Break here


if __name__ == '__main__':
    Call()
    print('TEST SUCEEDED!')



import sys
import unittest
import threading
import os
from _pydev_bundle.pydev_imports import SimpleXMLRPCServer
from _pydev_bundle.pydev_localhost import get_localhost
from _pydev_bundle.pydev_console_utils import StdIn
import socket
import time
from _pydevd_bundle import pydevd_io
import pytest


def eq_(a, b):
    if a != b:
        raise AssertionError('%s != %s' % (a, b))


try:
    from IPython import core
    has_ipython = True
except:
    has_ipython = False


@pytest.mark.skipif(not has_ipython, reason='IPython not available')
class TestBase(unittest.TestCase):

    def setUp(self):
        from _pydev_bundle.pydev_ipython_console_011 import get_pydev_frontend

        # PyDevFrontEnd depends on singleton in IPython, so you
        # can't make multiple versions. So we reuse self.front_end for
        # all the tests
        self.front_end = get_pydev_frontend(get_localhost(), 0)

        from pydev_ipython.inputhook import set_return_control_callback
        set_return_control_callback(lambda:True)
        self.front_end.clear_buffer()

    def tearDown(self):
        pass

    def add_exec(self, code, expected_more=False):
        more = self.front_end.add_exec(code)
        eq_(expected_more, more)

    def redirect_stdout(self):
        from IPython.utils import io

        self.original_stdout = sys.stdout
        sys.stdout = io.stdout = pydevd_io.IOBuf()

    def restore_stdout(self):
        from IPython.utils import io
        io.stdout = sys.stdout = self.original_stdout


@pytest.mark.skipif(not has_ipython, reason='IPython not available')
class TestPyDevFrontEnd(TestBase):

    def testAddExec_1(self):
        self.add_exec('if True:', True)

    def testAddExec_2(self):
        # Change: 'more' must now be controlled in the client side after the initial 'True' returned.
        self.add_exec('if True:\n    testAddExec_a = 10\n', False)
        assert 'testAddExec_a' in self.front_end.get_namespace()

    def testAddExec_3(self):
        assert 'testAddExec_x' not in self.front_end.get_namespace()
        self.add_exec('if True:\n    testAddExec_x = 10\n\n')
        assert 'testAddExec_x' in self.front_end.get_namespace()
        eq_(self.front_end.get_namespace()['testAddExec_x'], 10)

    def test_get_namespace(self):
        assert 'testGetNamespace_a' not in self.front_end.get_namespace()
        self.add_exec('testGetNamespace_a = 10')
        assert 'testGetNamespace_a' in self.front_end.get_namespace()
        eq_(self.front_end.get_namespace()['testGetNamespace_a'], 10)

    def test_complete(self):
        unused_text, matches = self.front_end.complete('%')
        assert len(matches) > 1, 'at least one magic should appear in completions'

    def test_complete_does_not_do_python_matches(self):
        # Test that IPython's completions do not do the things that
        # PyDev's completions will handle
        self.add_exec('testComplete_a = 5')
        self.add_exec('testComplete_b = 10')
        self.add_exec('testComplete_c = 15')
        unused_text, matches = self.front_end.complete('testComplete_')
        assert len(matches) == 0

    def testGetCompletions_1(self):
        # Test the merged completions include the standard completions
        self.add_exec('testComplete_a = 5')
        self.add_exec('testComplete_b = 10')
        self.add_exec('testComplete_c = 15')
        res = self.front_end.getCompletions('testComplete_', 'testComplete_')
        matches = [f[0] for f in res]
        assert len(matches) == 3
        eq_(set(['testComplete_a', 'testComplete_b', 'testComplete_c']), set(matches))

    def testGetCompletions_2(self):
        # Test that we get IPython completions in results
        # we do this by checking kw completion which PyDev does
        # not do by default
        self.add_exec('def ccc(ABC=123): pass')
        res = self.front_end.getCompletions('ccc(', '')
        matches = [f[0] for f in res]
        assert 'ABC=' in matches

    def testGetCompletions_3(self):
        # Test that magics return IPYTHON magic as type
        res = self.front_end.getCompletions('%cd', '%cd')
        assert len(res) == 1
        eq_(res[0][3], '12')  # '12' == IToken.TYPE_IPYTHON_MAGIC
        assert len(res[0][1]) > 100, 'docstring for %cd should be a reasonably long string'


@pytest.mark.skipif(not has_ipython, reason='IPython not available')
class TestRunningCode(TestBase):

    def test_print(self):
        self.redirect_stdout()
        try:
            self.add_exec('print("output")')
            eq_(sys.stdout.getvalue(), 'output\n')
        finally:
            self.restore_stdout()

    def testQuestionMark_1(self):
        self.redirect_stdout()
        try:
            self.add_exec('?')
            found = sys.stdout.getvalue()
            if len(found) < 1000:
                raise AssertionError('Expected IPython help to be big. Found: %s' % (found,))
        finally:
            self.restore_stdout()

    def testQuestionMark_2(self):
        self.redirect_stdout()
        try:
            self.add_exec('int?')
            found = sys.stdout.getvalue()
            if 'Convert' not in found:
                raise AssertionError('Expected to find "Convert" in %s' % (found,))
        finally:
            self.restore_stdout()

    def test_gui(self):
        try:
            import Tkinter
        except:
            return
        else:
            from pydev_ipython.inputhook import get_inputhook
            assert get_inputhook() is None
            self.add_exec('%gui tk')
            # we can't test the GUI works here because we aren't connected to XML-RPC so
            # nowhere for hook to run
            assert get_inputhook() is not None
            self.add_exec('%gui none')
            assert get_inputhook() is None

    def test_history(self):
        ''' Make sure commands are added to IPython's history '''
        self.redirect_stdout()
        try:
            self.add_exec('a=1')
            self.add_exec('b=2')
            _ih = self.front_end.get_namespace()['_ih']
            eq_(_ih[-1], 'b=2')
            eq_(_ih[-2], 'a=1')

            self.add_exec('history')
            hist = sys.stdout.getvalue().split('\n')
            eq_(hist[-1], '')
            eq_(hist[-2], 'history')
            eq_(hist[-3], 'b=2')
            eq_(hist[-4], 'a=1')
        finally:
            self.restore_stdout()

    def test_edit(self):
        ''' Make sure we can issue an edit command'''
        if os.environ.get('TRAVIS') == 'true':
            # This test is too flaky on travis.
            return

        from _pydev_bundle.pydev_ipython_console_011 import get_pydev_frontend

        called_RequestInput = [False]
        called_IPythonEditor = [False]

        def start_client_thread(client_port):

            class ClientThread(threading.Thread):

                def __init__(self, client_port):
                    threading.Thread.__init__(self)
                    self.client_port = client_port

                def run(self):

                    class HandleRequestInput:

                        def RequestInput(self):
                            called_RequestInput[0] = True
                            return '\n'

                        def IPythonEditor(self, name, line):
                            called_IPythonEditor[0] = (name, line)
                            return True

                    handle_request_input = HandleRequestInput()

                    from _pydev_bundle import pydev_localhost
                    self.client_server = client_server = SimpleXMLRPCServer(
                        (pydev_localhost.get_localhost(), self.client_port), logRequests=False)
                    client_server.register_function(handle_request_input.RequestInput)
                    client_server.register_function(handle_request_input.IPythonEditor)
                    client_server.serve_forever()

                def shutdown(self):
                    return
                    self.client_server.shutdown()

            client_thread = ClientThread(client_port)
            client_thread.daemon = True
            client_thread.start()
            return client_thread

        # PyDevFrontEnd depends on singleton in IPython, so you
        # can't make multiple versions. So we reuse self.front_end for
        # all the tests
        s = socket.socket()
        s.bind(('', 0))
        self.client_port = client_port = s.getsockname()[1]
        s.close()
        self.front_end = get_pydev_frontend(get_localhost(), client_port)

        client_thread = start_client_thread(self.client_port)
        orig_stdin = sys.stdin
        sys.stdin = StdIn(self, get_localhost(), self.client_port)
        try:
            filename = 'made_up_file.py'
            self.add_exec('%edit ' + filename)

            for i in range(10):
                if called_IPythonEditor[0] == (os.path.abspath(filename), '0'):
                    break
                time.sleep(.1)

            if not called_IPythonEditor[0]:
                #   File "/home/travis/miniconda/lib/python3.3/site-packages/IPython/core/interactiveshell.py", line 2883, in run_code
                #     exec(code_obj, self.user_global_ns, self.user_ns)
                #   File "<ipython-input-15-09583ca3bce1>", line 1, in <module>
                #     get_ipython().magic('edit made_up_file.py')
                #   File "/home/travis/miniconda/lib/python3.3/site-packages/IPython/core/interactiveshell.py", line 2205, in magic
                #     return self.run_line_magic(magic_name, magic_arg_s)
                #   File "/home/travis/miniconda/lib/python3.3/site-packages/IPython/core/interactiveshell.py", line 2126, in run_line_magic
                #     result = fn(*args,**kwargs)
                #   File "<string>", line 2, in edit
                #   File "/home/travis/miniconda/lib/python3.3/site-packages/IPython/core/magic.py", line 193, in <lambda>
                #     call = lambda f, *a, **k: f(*a, **k)
                #   File "/home/travis/miniconda/lib/python3.3/site-packages/IPython/core/magics/code.py", line 662, in edit
                #     self.shell.hooks.editor(filename,lineno)
                #   File "/home/travis/build/fabioz/PyDev.Debugger/pydev_ipython_console_011.py", line 70, in call_editor
                #     server.IPythonEditor(filename, str(line))
                #   File "/home/travis/miniconda/lib/python3.3/xmlrpc/client.py", line 1090, in __call__
                #     return self.__send(self.__name, args)
                #   File "/home/travis/miniconda/lib/python3.3/xmlrpc/client.py", line 1419, in __request
                #     verbose=self.__verbose
                #   File "/home/travis/miniconda/lib/python3.3/xmlrpc/client.py", line 1132, in request
                #     return self.single_request(host, handler, request_body, verbose)
                #   File "/home/travis/miniconda/lib/python3.3/xmlrpc/client.py", line 1143, in single_request
                #     http_conn = self.send_request(host, handler, request_body, verbose)
                #   File "/home/travis/miniconda/lib/python3.3/xmlrpc/client.py", line 1255, in send_request
                #     self.send_content(connection, request_body)
                #   File "/home/travis/miniconda/lib/python3.3/xmlrpc/client.py", line 1285, in send_content
                #     connection.endheaders(request_body)
                #   File "/home/travis/miniconda/lib/python3.3/http/client.py", line 1061, in endheaders
                #     self._send_output(message_body)
                #   File "/home/travis/miniconda/lib/python3.3/http/client.py", line 906, in _send_output
                #     self.send(msg)
                #   File "/home/travis/miniconda/lib/python3.3/http/client.py", line 844, in send
                #     self.connect()
                #   File "/home/travis/miniconda/lib/python3.3/http/client.py", line 822, in connect
                #     self.timeout, self.source_address)
                #   File "/home/travis/miniconda/lib/python3.3/socket.py", line 435, in create_connection
                #     raise err
                #   File "/home/travis/miniconda/lib/python3.3/socket.py", line 426, in create_connection
                #     sock.connect(sa)
                # ConnectionRefusedError: [Errno 111] Connection refused

                # I.e.: just warn that the test failing, don't actually fail.
                sys.stderr.write('Test failed: this test is brittle in travis because sometimes the connection is refused (as above) and we do not have a callback.\n')
                return

            eq_(called_IPythonEditor[0], (os.path.abspath(filename), '0'))
            assert called_RequestInput[0], "Make sure the 'wait' parameter has been respected"
        finally:
            sys.stdin = orig_stdin
            client_thread.shutdown()




from robocorp_ls_core.robotframework_log import get_logger

log = get_logger(__name__)


class Callback(object):
    """
    Note that it's thread safe to register/unregister callbacks while callbacks
    are being notified, but it's not thread-safe to register/unregister at the
    same time in multiple threads.
    """

    def __init__(self):
        self._callbacks = []

    def register(self, callback):
        new_callbacks = self._callbacks[:]
        new_callbacks.append(callback)
        self._callbacks = new_callbacks

    def unregister(self, callback):
        new_callbacks = [x for x in self._callbacks if x != callback]
        self._callbacks = new_callbacks

    def __call__(self, *args, **kwargs):
        for c in self._callbacks:
            try:
                c(*args, **kwargs)
            except:
                log.exception("Error in callback.")


class CallbackWithReturn(Callback):
    """
    A callback that returns the first non-None value returned
    from a registered caller.
    """

    def __call__(self, *args, **kwargs):
        for c in self._callbacks:
            try:
                ret = c(*args, **kwargs)
            except:
                log.exception("Error in callback with return.")
                continue
            if ret is not None:
                return ret
        return None



# This is a dummy for code-completion purposes.

def __unicode__(self):
    """
    Return "app_label.model_label.manager_name". 
    """

def _copy_to_model(self, model):
    """
    Makes a copy of the manager and assigns it to 'model', which should be
    a child of the existing model (used when inheriting a manager from an
    abstract base class).
    """


def _db(self):
    """

    """


def _get_queryset_methods(cls, queryset_class):
    """

    """


def _hints(self):
    """
    dict() -> new empty dictionary
    dict(mapping) -> new dictionary initialized from a mapping object's
        (key, value) pairs
    dict(iterable) -> new dictionary initialized as if via:
        d = {}
        for k, v in iterable:
            d[k] = v
    dict(**kwargs) -> new dictionary initialized with the name=value pairs
        in the keyword argument list.  For example:  dict(one=1, two=2)
    """


def _inherited(self):
    """

    """


def _insert(self, *args, **kwargs):
    """
    Inserts a new record for the given model. This provides an interface to
    the InsertQuery class and is how Model.save() is implemented.
    """


def _queryset_class(self):
    """
    Represents a lazy database lookup for a set of objects.
    """


def _set_creation_counter(self):
    """
    Sets the creation counter value for this instance and increments the
    class-level copy.
    """


def _update(self, *args, **kwargs):
    """
    A version of update that accepts field objects instead of field names.
    Used primarily for model saving and not intended for use by general
    code (it requires too much poking around at model internals to be
    useful at that level).
    """


def aggregate(self, *args, **kwargs):
    """
    Returns a dictionary containing the calculations (aggregation)
    over the current queryset
    
    If args is present the expression is passed as a kwarg using
    the Aggregate object's default alias.
    """


def all(self):
    """
    @rtype: django.db.models.query.QuerySet
    """


def annotate(self, *args, **kwargs):
    """
    Return a query set in which the returned objects have been annotated
    with data aggregated from related fields.
    """


def bulk_create(self, *args, **kwargs):
    """
    Inserts each of the instances into the database. This does *not* call
    save() on each of the instances, does not send any pre/post save
    signals, and does not set the primary key attribute if it is an
    autoincrement field.
    """


def check(self, **kwargs):
    """

    """


def complex_filter(self, *args, **kwargs):
    """
    Returns a new QuerySet instance with filter_obj added to the filters.
    
    filter_obj can be a Q object (or anything with an add_to_query()
    method) or a dictionary of keyword lookup arguments.
    
    This exists to support framework features such as 'limit_choices_to',
    and usually it will be more natural to use other methods.
    
    @rtype: django.db.models.query.QuerySet
    """


def contribute_to_class(self, model, name):
    """

    """


def count(self, *args, **kwargs):
    """
    Performs a SELECT COUNT() and returns the number of records as an
    integer.
    
    If the QuerySet is already fully cached this simply returns the length
    of the cached results set to avoid multiple SELECT COUNT(*) calls.
    """


def create(self, *args, **kwargs):
    """
    Creates a new object with the given kwargs, saving it to the database
    and returning the created object.
    """


def creation_counter(self):
    """

    """


def dates(self, *args, **kwargs):
    """
    Returns a list of date objects representing all available dates for
    the given field_name, scoped to 'kind'.
    """


def datetimes(self, *args, **kwargs):
    """
    Returns a list of datetime objects representing all available
    datetimes for the given field_name, scoped to 'kind'.
    """


def db(self):
    """

    """


def db_manager(self, using=None, hints=None):
    """

    """


def defer(self, *args, **kwargs):
    """
    Defers the loading of data for certain fields until they are accessed.
    The set of fields to defer is added to any existing set of deferred
    fields. The only exception to this is if None is passed in as the only
    parameter, in which case all deferrals are removed (None acts as a
    reset option).
    """


def distinct(self, *args, **kwargs):
    """
    Returns a new QuerySet instance that will select only distinct results.
    
    @rtype: django.db.models.query.QuerySet
    """


def earliest(self, *args, **kwargs):
    """

    """


def exclude(self, *args, **kwargs):
    """
    Returns a new QuerySet instance with NOT (args) ANDed to the existing
    set.
    
    @rtype: django.db.models.query.QuerySet
    """


def exists(self, *args, **kwargs):
    """

    """


def extra(self, *args, **kwargs):
    """
    Adds extra SQL fragments to the query.
    """


def filter(self, *args, **kwargs):
    """
    Returns a new QuerySet instance with the args ANDed to the existing
    set.
    
    @rtype: django.db.models.query.QuerySet
    """


def first(self, *args, **kwargs):
    """
    Returns the first object of a query, returns None if no match is found.
    """


def from_queryset(cls, queryset_class, class_name=None):
    """

    """


def get(self, *args, **kwargs):
    """
    Performs the query and returns a single object matching the given
    keyword arguments.
    """


def get_or_create(self, *args, **kwargs):
    """
    Looks up an object with the given kwargs, creating one if necessary.
    Returns a tuple of (object, created), where created is a boolean
    specifying whether an object was created.
    """


def get_queryset(self):
    """
    Returns a new QuerySet object.  Subclasses can override this method to
    easily customize the behavior of the Manager.
    
    @rtype: django.db.models.query.QuerySet
    """


def in_bulk(self, *args, **kwargs):
    """
    Returns a dictionary mapping each of the given IDs to the object with
    that ID.
    """


def iterator(self, *args, **kwargs):
    """
    An iterator over the results from applying this QuerySet to the
    database.
    """


def last(self, *args, **kwargs):
    """
    Returns the last object of a query, returns None if no match is found.
    """


def latest(self, *args, **kwargs):
    """

    """


def model(self):
    """
    MyModel(id)
    """


def none(self, *args, **kwargs):
    """
    Returns an empty QuerySet.
    
    @rtype: django.db.models.query.QuerySet
    """


def only(self, *args, **kwargs):
    """
    Essentially, the opposite of defer. Only the fields passed into this
    method and that are not already specified as deferred are loaded
    immediately when the queryset is evaluated.
    """


def order_by(self, *args, **kwargs):
    """
    Returns a new QuerySet instance with the ordering changed.
    
    @rtype: django.db.models.query.QuerySet
    """


def prefetch_related(self, *args, **kwargs):
    """
    Returns a new QuerySet instance that will prefetch the specified
    Many-To-One and Many-To-Many related objects when the QuerySet is
    evaluated.
    
    When prefetch_related() is called more than once, the list of lookups to
    prefetch is appended to. If prefetch_related(None) is called, the list
    is cleared.
    
    @rtype: django.db.models.query.QuerySet
    """


def raw(self, *args, **kwargs):
    """

    """


def reverse(self, *args, **kwargs):
    """
    Reverses the ordering of the QuerySet.
    
    @rtype: django.db.models.query.QuerySet
    """


def select_for_update(self, *args, **kwargs):
    """
    Returns a new QuerySet instance that will select objects with a
    FOR UPDATE lock.
    
    @rtype: django.db.models.query.QuerySet
    """


def select_related(self, *args, **kwargs):
    """
    Returns a new QuerySet instance that will select related objects.
    
    If fields are specified, they must be ForeignKey fields and only those
    related objects are included in the selection.
    
    If select_related(None) is called, the list is cleared.
    
    @rtype: django.db.models.query.QuerySet
    """


def update(self, *args, **kwargs):
    """
    Updates all elements in the current QuerySet, setting all the given
    fields to the appropriate values.
    """


def update_or_create(self, *args, **kwargs):
    """
    Looks up an object with the given kwargs, updating one with defaults
    if it exists, otherwise creates a new one.
    Returns a tuple (object, created), where created is a boolean
    specifying whether an object was created.
    """


def using(self, *args, **kwargs):
    """
    Selects which database this QuerySet should execute its query against.
    
    @rtype: django.db.models.query.QuerySet
    """


def values(self, *args, **kwargs):
    """

    """


def values_list(self, *args, **kwargs):
    """

    """




"""
Errors checkers
"""
import re

from robot.api import Token

try:
    from robot.api.parsing import If
except ImportError:
    If = None

from robocop.checkers import VisitorChecker
from robocop.rules import Rule, RuleSeverity
from robocop.utils import ROBOT_VERSION, find_robot_vars, token_col

rules = {
    "0401": Rule(
        rule_id="0401",
        name="parsing-error",
        msg="Robot Framework syntax error: {{ error_msg }}",
        severity=RuleSeverity.ERROR,
    ),
    "0402": Rule(
        rule_id="0402",
        name="not-enough-whitespace-after-setting",
        msg="Provide at least two spaces after '{{ setting_name }}' setting",
        severity=RuleSeverity.ERROR,
        docs="""
        Example of rule violation::

            *** Test Cases ***
            Test
                [Documentation] doc  # only one space after [Documentation]
                Keyword
                
            *** Keywords ***
            Keyword
                [Documentation]  This is doc
                [Arguments] ${var}  # only one space after [Arguments]
                Should Be True  ${var}
            
        """,
    ),
    "0403": Rule(
        rule_id="0403",
        name="missing-keyword-name",
        msg="Missing keyword name when calling some values",
        severity=RuleSeverity.ERROR,
        docs="""
        Example of rule violation::
        
            *** Keywords ***
            Keyword
                ${var}
                ${one}      ${two}

        """,
    ),
    "0404": Rule(
        rule_id="0404",
        name="variables-import-with-args",
        msg="Robot and YAML variable files do not take arguments",
        severity=RuleSeverity.ERROR,
        docs="""
        Example of rule violation::
        
            *** Settings ***
            Variables    vars.yaml    arg1
            Variables    variables.robot    arg
        
        """,
    ),
    "0405": Rule(
        rule_id="0405",
        name="invalid-continuation-mark",
        msg="Invalid continuation mark '{{ mark }}'. It should be '...'",
        severity=RuleSeverity.ERROR,
        docs="""
        Example of rule violation::
        
            Keyword
            ..  ${var}  # .. instead of ...
            ...  1
            ....  2  # .... instead of ...

        """,
    ),
    # there is not-enough-whitespace-after-newline-marker for keyword calls already
    "0406": Rule(
        rule_id="0406",
        name="not-enough-whitespace-after-newline-marker",
        msg="Provide at least two spaces after '...' marker",
        severity=RuleSeverity.ERROR,
        docs="""
        Example of rule violation::
        
            @{LIST}  1
            ... 2  # not enough whitespace
            ...  3

        """,
    ),
    "0407": Rule(
        rule_id="0407",
        name="invalid-argument",
        msg="{{ error_msg }}",
        severity=RuleSeverity.ERROR,
        version=">=4.0",
        docs="""
        Argument names should follow variable naming syntax: start with identifier (`$`, `@` or `&`) and enclosed in 
        curly brackets (`{}`).
        
        Valid names::
        
            Keyword
                [Arguments]    ${var}    @{args}    &{config}    ${var}=default
        
        Invalid names::
        
            Keyword
                [Arguments]    {var}    @args}    var=default
        
        """,
    ),
    "0408": Rule(
        rule_id="0408",
        name="non-existing-setting",
        msg="{{ error_msg }}",
        severity=RuleSeverity.ERROR,
        docs="""
        Non-existing setting can't be used in the code.
        
        Rule violation example::
        
           *** Test Case ***
               [Not Existing]  arg
               [Arguments]  ${arg}
    
        """,
    ),
    "0409": Rule(
        rule_id="0409",
        name="setting-not-supported",
        msg="Setting '[{{ setting_name }}]' is not supported in {{ test_or_keyword }}. "
        "Allowed are: {{ allowed_settings }}",
        severity=RuleSeverity.ERROR,
        docs="""
        Following settings are supported in Test Case::
        
            [Documentation]	 Used for specifying a test case documentation.
            [Tags]	         Used for tagging test cases.
            [Setup]	         Used for specifying a test setup.
            [Teardown]	     Used for specifying a test teardown.
            [Template]	     Used for specifying a template keyword.
            [Timeout]	     Used for specifying a test case timeout.
        
        Following settings are supported in Keyword::
        
            [Documentation]	 Used for specifying a user keyword documentation.
            [Tags]	         Used for specifying user keyword tags.
            [Arguments]	     Used for specifying user keyword arguments.
            [Return]	     Used for specifying user keyword return values.
            [Teardown]	     Used for specifying user keyword teardown.
            [Timeout]	     Used for specifying a user keyword timeout.
        
        """,
    ),
    "0410": Rule(
        rule_id="0410",
        name="not-enough-whitespace-after-variable",
        msg="Provide at least two spaces after '{{ variable_name }}' variable name",
        severity=RuleSeverity.ERROR,
        version=">=4.0",
        docs="""
        Example of rule violation::
        
            ${variable} 1  # not enough whitespace
            ${other_var}  2
        
        """,
    ),
    "0411": Rule(
        rule_id="0411",
        name="not-enough-whitespace-after-suite-setting",
        msg="Provide at least two spaces after '{{ setting_name }}' setting",
        severity=RuleSeverity.ERROR,
        docs="""
        Example of rule violation::
        
            *** Settings ***
            Library Collections  # not enough whitespace
            Force Tags  tag
            ...  tag2
            Suite Setup Keyword  # not enough whitespace
        
        """,
    ),
    "0412": Rule(
        rule_id="0412",
        name="invalid-for-loop",
        msg="Invalid for loop syntax: {{ error_msg }}",
        severity=RuleSeverity.ERROR,
        version=">=4.0",
    ),
    "0413": Rule(
        rule_id="0413",
        name="invalid-if",
        msg="Invalid IF syntax: {{ error_msg }}",
        severity=RuleSeverity.ERROR,
        version=">=4.0",
    ),
    "0414": Rule(
        rule_id="0414",
        name="return-in-test-case",
        msg="RETURN can only be used inside a user keyword",
        severity=RuleSeverity.ERROR,
        version=">=5.0",
    ),
}


class ParsingErrorChecker(VisitorChecker):
    """Checker that parses Robot Framework DataErrors."""

    reports = (
        "parsing-error",
        "invalid-continuation-mark",
        "not-enough-whitespace-after-newline-marker",
        "invalid-argument",
        "non-existing-setting",
        "setting-not-supported",
        "not-enough-whitespace-after-variable",
        "not-enough-whitespace-after-suite-setting",
        "invalid-for-loop",
        "invalid-if",
        "return-in-test-case",
    )

    keyword_only_settings = {"Arguments", "Return"}
    keyword_settings = [
        "[Documentation]",
        "[Tags]",
        "[Arguments]",
        "[Return]",
        "[Teardown]",
        "[Timeout]",
    ]
    test_case_only_settings = {"Setup", "Template"}
    test_case_settings = [
        "[Documentation]",
        "[Tags]",
        "[Setup]",
        "[Teardown]",
        "[Template]",
        "[Timeout]",
    ]
    suite_settings = {
        "documentation": "Documentation",
        "suitesetup": "Suite Setup",
        "suiteteardown": "Suite Teardown",
        "metadata": "Metadata",
        "testsetup": "Test Setup",
        "testteardown": "Test Teardown",
        "testtemplate": "Test Template",
        "testtimeout": "Test Timeout",
        "forcetags": "Force Tags",
        "defaulttags": "Default Tags",
        "library": "Library",
        "resource": "Resource",
        "variables": "Variables",
    }
    ignore_errors = ("can only be used inside a loop",)

    def __init__(self):
        super().__init__()
        self.in_block = None

    def visit_File(self, node):
        self.generic_visit(node)

    def visit_If(self, node):  # noqa
        self.in_block = node  # to ensure we're in IF for `invalid-if` rule
        self.parse_errors(node)
        self.generic_visit(node)

    visit_For = visit_While = visit_Try = visit_If

    def visit_KeywordCall(self, node):  # noqa
        if node.keyword and node.keyword.startswith("..."):
            self.report("not-enough-whitespace-after-newline-marker", node=node)
        self.generic_visit(node)

    def visit_Statement(self, node):  # noqa
        self.parse_errors(node)

    def parse_errors(self, node):  # noqa
        if node is None:
            return
        if ROBOT_VERSION.major != 3:
            for index, error in enumerate(node.errors):
                self.handle_error(node, error, error_index=index)
        else:
            self.handle_error(node, node.error)

    def handle_error(self, node, error, error_index=0):  # noqa
        if not error:
            return
        if any(should_ignore in error for should_ignore in self.ignore_errors):
            return
        if "Invalid argument syntax" in error:
            self.handle_invalid_syntax(node, error)
        elif "Non-existing setting" in error:
            self.handle_invalid_setting(node, error)
        elif "Invalid variable name" in error:
            self.handle_invalid_variable(node, error)
        elif "RETURN can only be used inside" in error:
            self.report("return-in-test-case", node=node, col=token_col(node, "RETURN STATEMENT"))
        elif "IF" in error or ("ELSE" in error and If and isinstance(self.in_block, If)):
            self.handle_invalid_block(node, error, "invalid-if")
        elif "FOR loop" in error:
            self.handle_invalid_block(node, error, "invalid-for-loop")
        elif "Non-default argument after default arguments" in error or "Only last argument can be kwargs" in error:
            self.handle_positional_after_named(node, error_index)
        elif "is allowed only once. Only the first value is used" in error:
            return
        else:
            error = error.replace("\n   ", "")
            self.report("parsing-error", error_msg=error, node=node)

    def handle_invalid_block(self, node, error, block_name):
        if hasattr(node, "header"):
            token = node.header.get_token(node.header.type)
        else:
            token = node.get_token(node.type)
        self.report(
            block_name,
            error_msg=error.replace("Robot Framework syntax error: ", "")[:-1],
            node=token,
            col=token.col_offset + 1,
        )

    def handle_invalid_syntax(self, node, error):
        # robot doesn't report on exact token, so we need to find it
        match = re.search("'(.+)'", error)
        if not match:
            return
        for arg in node.get_tokens(Token.ARGUMENT):
            value, *_ = arg.value.split("=", maxsplit=1)
            if value == match.group(1):
                self.report("invalid-argument", error_msg=error[:-1], node=arg, col=arg.col_offset + 1)
                return
        self.report("parsing-error", error_msg=error, node=node)

    def handle_invalid_setting(self, node, error):
        setting_error = re.search("Non-existing setting '(.*)'.", error)
        if not setting_error:
            return
        setting_error = setting_error.group(1)
        if not setting_error:
            return
        if setting_error.lstrip().startswith(".."):
            self.handle_invalid_continuation_mark(node, node.data_tokens[0].value)
        elif setting_error in self.keyword_only_settings:
            self.report(
                "setting-not-supported",
                setting_name=setting_error,
                test_or_keyword="Test Case",
                allowed_settings=", ".join(self.test_case_settings),
                node=node,
            )
        elif setting_error in self.test_case_only_settings:
            self.report(
                "setting-not-supported",
                setting_name=setting_error,
                test_or_keyword="Keyword",
                allowed_settings=", ".join(self.keyword_settings),
                node=node,
            )
        else:
            suite_sett_cand = setting_error.replace(" ", "").lower()
            for setting in self.suite_settings:
                if suite_sett_cand.startswith(setting):
                    if setting_error[0].strip():  # filter out "suite-setting-should-be-left-aligned"
                        self.report(
                            "not-enough-whitespace-after-suite-setting",
                            setting_name=self.suite_settings[setting],
                            node=node,
                        )
                    return
            error = error.replace("\n   ", "").replace("Robot Framework syntax error: ", "")
            if error.endswith("."):
                error = error[:-1]
            self.report("non-existing-setting", error_msg=error, node=node)

    def handle_invalid_variable(self, node, error):
        var_error = re.search("Invalid variable name '(.*)'.", error)
        if not var_error or not var_error.group(1):  # empty variable name due to invalid parsing
            return
        elif var_error.group(1).lstrip().startswith(".."):
            self.handle_invalid_continuation_mark(node, var_error.group(1))
        elif not var_error.group(1)[0].strip():  # not left aligned variable
            return
        else:
            variable_token = node.get_token(Token.VARIABLE)
            variables = find_robot_vars(variable_token.value) if variable_token else None
            if variables and variables[0][0] == 0:
                self.report(
                    "not-enough-whitespace-after-variable",
                    variable_name=variable_token.value,
                    node=variable_token,
                    col=variable_token.col_offset + 1,
                )
            else:
                error = error.replace("\n   ", "")
                self.report("parsing-error", error_msg=error, node=node)

    def handle_invalid_continuation_mark(self, node, name):
        stripped = name.lstrip()
        if len(stripped) == 2 or not stripped[2].strip():
            self.report("invalid-continuation-mark", mark=stripped, node=node, col=name.find(".") + 1)
        elif len(stripped) >= 4:
            if stripped[:4] == "....":
                self.report("invalid-continuation-mark", mark=stripped, node=node, col=name.find(".") + 1)
            else:  # '... ' or '...value' or '...\t'
                self.report(
                    "not-enough-whitespace-after-newline-marker",
                    node=node,
                    col=name.find(".") + 1,
                )

    @staticmethod
    def is_var_positional(value):
        if not value:
            return False
        if value.startswith("&") or "=" in value:
            return True
        return False

    def handle_positional_after_named(self, node, error_index):
        """
        Robot Framework reports all errors on parent node. That's why we need to find which token is invalid - and in
        case there are several invalid tokens we need to skip tokens that were already reported for particular node.
        """
        named_found = False
        token = node
        skip = error_index
        for token in node.get_tokens(Token.ARGUMENT):
            if named_found and not self.is_var_positional(token.value):
                if not skip:
                    break
                skip -= 1
            named_found = self.is_var_positional(token.value)
        self.report(
            "parsing-error",
            error_msg=f"Positional argument '{token.value}' follows named argument",
            node=token,
            col=token.col_offset + 1,
        )


class TwoSpacesAfterSettingsChecker(VisitorChecker):
    """Checker for not enough whitespaces after [Setting] header."""

    reports = ("not-enough-whitespace-after-setting",)

    def __init__(self):
        self.headers = {
            "arguments",
            "documentation",
            "setup",
            "timeout",
            "teardown",
            "template",
            "tags",
        }
        self.setting_pattern = re.compile(r"\[\s?(\w+)\s?\]")
        super().__init__()

    def visit_KeywordCall(self, node):  # noqa
        """Invalid settings like '[Arguments] ${var}' will be parsed as keyword call"""
        if not node.keyword:
            return

        match = self.setting_pattern.match(node.keyword)
        if not match:
            return
        if match.group(1).lower() in self.headers:
            self.report(
                "not-enough-whitespace-after-setting",
                setting_name=match.group(0),
                node=node,
                col=node.data_tokens[0].col_offset + 1,
            )


class MissingKeywordName(VisitorChecker):
    """Checker for missing keyword name."""

    reports = ("missing-keyword-name",)

    def visit_EmptyLine(self, node):  # noqa
        if ROBOT_VERSION.major < 5:
            return
        assign_token = node.get_token(Token.ASSIGN)
        if assign_token:
            self.report(
                "missing-keyword-name",
                node=node,
                lineno=node.lineno,
                col=assign_token.col_offset + 1,
            )

    def visit_KeywordCall(self, node):  # noqa
        if node.keyword is None:
            self.report(
                "missing-keyword-name",
                node=node,
                lineno=node.lineno,
                col=node.data_tokens[0].col_offset + 1,
            )


class VariablesImportErrorChecker(VisitorChecker):
    """Checker for syntax error in variables import."""

    reports = ("variables-import-with-args",)

    def visit_VariablesImport(self, node):  # noqa
        if node.name and not node.name.endswith(".py") and node.get_token(Token.ARGUMENT):
            self.report("variables-import-with-args", node=node)



import os
import sys
import platform

TEST_CYTHON = os.getenv('PYDEVD_USE_CYTHON', None) == 'YES'
PYDEVD_TEST_VM = os.getenv('PYDEVD_TEST_VM', None)

IS_PY36_OR_GREATER = sys.version_info[0:2] >= (3, 6)
IS_CPYTHON = platform.python_implementation() == 'CPython'

IS_PY36 = False
if sys.version_info[0] == 3 and sys.version_info[1] == 6:
    IS_PY36 = True

TEST_DJANGO = False
TEST_FLASK = False
TEST_CHERRYPY = False
TEST_GEVENT = False

try:
    import django
    TEST_DJANGO = True
except:
    pass

try:
    import flask
    TEST_FLASK = True
except:
    pass

try:
    import cherrypy
    TEST_CHERRYPY = True
except:
    pass

try:
    import gevent
    TEST_GEVENT = True
except:
    pass



from robocorp_ls_core.config import Config
from robotframework_ls.impl.robot_lsp_constants import ALL_ROBOT_OPTIONS
import os.path


class RobotConfig(Config):
    ALL_OPTIONS = ALL_ROBOT_OPTIONS


def get_robotframework_ls_home():
    user_home = os.getenv("ROBOTFRAMEWORK_LS_USER_HOME", None)
    if user_home is None:
        user_home = os.path.expanduser("~")

    return os.path.join(user_home, ".robotframework-ls")


def create_convert_keyword_format_func(config):
    if config is not None:
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT,
        )

        keyword_format = config.get_setting(
            OPTION_ROBOT_COMPLETION_KEYWORDS_FORMAT, str, ""
        )

        if keyword_format:
            keyword_format = keyword_format.lower().replace(" ", "_").strip()
            # Convert its format depending on
            # the user configuration.
            if keyword_format == "first_upper":
                return lambda label: label.capitalize()

            elif keyword_format == "title_case":
                return lambda label: label.title()

            elif keyword_format == "all_lower":
                return lambda label: label.lower()

            elif keyword_format == "all_upper":
                return lambda label: label.upper()

    return lambda x: x



#!/usr/bin/env python
"""Simple pyglet example to manually test event loop integration.

To run this:
1) Enable the PyDev GUI event loop integration for pyglet
2) do an execfile on this script
3) ensure you have a working GUI simultaneously with an
   interactive console
"""

if __name__ == '__main__':
    import pyglet
    
    
    window = pyglet.window.Window()
    label = pyglet.text.Label('Hello, world',
                              font_name='Times New Roman',
                              font_size=36,
                              x=window.width//2, y=window.height//2,
                              anchor_x='center', anchor_y='center')
    @window.event
    def on_close():
        window.close()
    
    @window.event
    def on_draw():
        window.clear()
        label.draw()



from robot.parsing.model.statements import Statement
import ast


class _Visitor(ast.NodeVisitor):
    def __init__(self, model):
        self.text = []
        self.visit(model)

    def visit(self, node):
        if isinstance(node, Statement):
            visitor = self.visit_Statement
        else:
            visitor = self.generic_visit
        visitor(node)

    def visit_Statement(self, node):  # noqa
        for token in node.tokens:
            self.text.append(token.value)


def ast_to_code(node):
    v = _Visitor(node)
    for i, txt in enumerate(reversed(v.text)):
        if not txt.strip():
            continue
        else:
            break

    if i > 0:
        v.text = v.text[:-i]
        v.text.append("\n")

    return "".join(v.text)



from robocorp_ls_core.protocols import IDocument
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import IVariablesCollector
from robotframework_ls.impl.variable_types import VariableFoundFromPythonAst


log = get_logger(__name__)


def _gen_var_from_python_ast(variable_import_doc, collector, value_node, target_node):
    try:
        import ast as ast_module

        varname = None
        if isinstance(target_node, (ast_module.Constant, ast_module.Str)):
            varname = target_node.s

        elif isinstance(target_node, ast_module.Name):
            varname = target_node.id

        if varname is not None:
            varname = str(varname)
            if collector.accepts(varname):
                value = ""
                try:
                    # Only available for Python 3.8 onwards...
                    end_lineno = getattr(value_node, "end_lineno", None)
                    if end_lineno is None:
                        end_lineno = value_node.lineno

                    # Only available for Python 3.8 onwards...
                    end_col_offset = getattr(value_node, "end_col_offset", None)
                    if end_col_offset is None:
                        end_col_offset = 99999999
                    value = variable_import_doc.get_range(
                        value_node.lineno - 1,
                        value_node.col_offset,
                        end_lineno - 1,
                        end_col_offset,
                    )
                except:
                    log.exception("Error dealing with value node: %s", value_node)

                variable_found = VariableFoundFromPythonAst(
                    variable_import_doc.path,
                    target_node.lineno - 1,
                    target_node.col_offset,
                    target_node.lineno - 1,
                    target_node.col_offset + len(varname),
                    value,
                    variable_name=varname,
                )
                collector.on_variable(variable_found)
    except:
        log.exception(
            "Error dealing with target node: %s, value node: %s",
            target_node,
            value_node,
        )


def collect_variables_from_python_ast(
    python_ast, python_doc: IDocument, collector: IVariablesCollector
):
    import ast as ast_module

    try:
        for node in python_ast.body:
            if isinstance(node, ast_module.AnnAssign):
                _gen_var_from_python_ast(python_doc, collector, node.value, node.target)

            elif isinstance(node, ast_module.Assign):
                for target in node.targets:
                    _gen_var_from_python_ast(python_doc, collector, node.value, target)

            elif isinstance(node, ast_module.FunctionDef):
                if node.name in ("get_variables", "getVariables"):
                    # https://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#getting-variables-from-a-special-function
                    for b in node.body:
                        if isinstance(b, ast_module.Return):
                            if isinstance(b.value, ast_module.Dict):
                                for key, value in zip(b.value.keys, b.value.values):
                                    _gen_var_from_python_ast(
                                        python_doc, collector, value, key
                                    )
    except:
        log.exception("Error collecting variables from Python AST.")



# encoding: utf-8
"""
This script tests utility functions.
"""

import errno
import os
import os.path
import shutil
import sys
import tempfile
import unittest

from pathspec.util import iter_tree_entries, iter_tree_files, RecursionError, normalize_file


class IterTreeTest(unittest.TestCase):
	"""
	The ``IterTreeTest`` class tests `pathspec.util.iter_tree_files()`.
	"""

	def make_dirs(self, dirs):
		"""
		Create the specified directories.
		"""
		for dir in dirs:
			os.mkdir(os.path.join(self.temp_dir, self.ospath(dir)))

	def make_files(self, files):
		"""
		Create the specified files.
		"""
		for file in files:
			self.mkfile(os.path.join(self.temp_dir, self.ospath(file)))

	def make_links(self, links):
		"""
		Create the specified links.
		"""
		for link, node in links:
			os.symlink(os.path.join(self.temp_dir, self.ospath(node)), os.path.join(self.temp_dir, self.ospath(link)))

	@staticmethod
	def mkfile(file):
		"""
		Creates an empty file.
		"""
		with open(file, 'wb'):
			pass

	@staticmethod
	def ospath(path):
		"""
		Convert the POSIX path to a native OS path.
		"""
		return os.path.join(*path.split('/'))

	def require_realpath(self):
		"""
		Skips the test if `os.path.realpath` does not properly support
		symlinks.
		"""
		if self.broken_realpath:
			raise unittest.SkipTest("`os.path.realpath` is broken.")

	def require_symlink(self):
		"""
		Skips the test if `os.symlink` is not supported.
		"""
		if self.no_symlink:
			raise unittest.SkipTest("`os.symlink` is not supported.")

	def setUp(self):
		"""
		Called before each test.
		"""
		self.temp_dir = tempfile.mkdtemp()

	def tearDown(self):
		"""
		Called after each test.
		"""
		shutil.rmtree(self.temp_dir)

	def test_1_files(self):
		"""
		Tests to make sure all files are found.
		"""
		self.make_dirs([
			'Empty',
			'Dir',
			'Dir/Inner',
		])
		self.make_files([
			'a',
			'b',
			'Dir/c',
			'Dir/d',
			'Dir/Inner/e',
			'Dir/Inner/f',
		])
		results = set(iter_tree_files(self.temp_dir))
		self.assertEqual(results, set(map(self.ospath, [
			'a',
			'b',
			'Dir/c',
			'Dir/d',
			'Dir/Inner/e',
			'Dir/Inner/f',
		])))

	def test_2_0_check_symlink(self):
		"""
		Tests whether links can be created.
		"""
		# NOTE: Windows does not support `os.symlink` for Python 2. Windows
		# Vista and greater supports `os.symlink` for Python 3.2+.
		no_symlink = None
		try:
			file = os.path.join(self.temp_dir, 'file')
			link = os.path.join(self.temp_dir, 'link')
			self.mkfile(file)

			try:
				os.symlink(file, link)
			except (AttributeError, NotImplementedError):
				no_symlink = True
				raise
			no_symlink = False

		finally:
			self.__class__.no_symlink = no_symlink

	def test_2_1_check_realpath(self):
		"""
		Tests whether `os.path.realpath` works properly with symlinks.
		"""
		# NOTE: Windows does not follow symlinks with `os.path.realpath`
		# which is what we use to detect recursion. See <https://bugs.python.org/issue9949>
		# for details.
		broken_realpath = None
		try:
			self.require_symlink()
			file = os.path.join(self.temp_dir, 'file')
			link = os.path.join(self.temp_dir, 'link')
			self.mkfile(file)
			os.symlink(file, link)

			try:
				self.assertEqual(os.path.realpath(file), os.path.realpath(link))
			except AssertionError:
				broken_realpath = True
				raise
			broken_realpath = False

		finally:
			self.__class__.broken_realpath = broken_realpath

	def test_2_2_links(self):
		"""
		Tests to make sure links to directories and files work.
		"""
		self.require_symlink()
		self.make_dirs([
			'Dir',
		])
		self.make_files([
			'a',
			'b',
			'Dir/c',
			'Dir/d',
		])
		self.make_links([
			('ax', 'a'),
			('bx', 'b'),
			('Dir/cx', 'Dir/c'),
			('Dir/dx', 'Dir/d'),
			('DirX', 'Dir'),
		])
		results = set(iter_tree_files(self.temp_dir))
		self.assertEqual(results, set(map(self.ospath, [
			'a',
			'ax',
			'b',
			'bx',
			'Dir/c',
			'Dir/cx',
			'Dir/d',
			'Dir/dx',
			'DirX/c',
			'DirX/cx',
			'DirX/d',
			'DirX/dx',
		])))

	def test_2_3_sideways_links(self):
		"""
		Tests to make sure the same directory can be encountered multiple
		times via links.
		"""
		self.require_symlink()
		self.make_dirs([
			'Dir',
			'Dir/Target',
		])
		self.make_files([
			'Dir/Target/file',
		])
		self.make_links([
			('Ax', 'Dir'),
			('Bx', 'Dir'),
			('Cx', 'Dir/Target'),
			('Dx', 'Dir/Target'),
			('Dir/Ex', 'Dir/Target'),
			('Dir/Fx', 'Dir/Target'),
		])
		results = set(iter_tree_files(self.temp_dir))
		self.assertEqual(results, set(map(self.ospath, [
			'Ax/Ex/file',
			'Ax/Fx/file',
			'Ax/Target/file',
			'Bx/Ex/file',
			'Bx/Fx/file',
			'Bx/Target/file',
			'Cx/file',
			'Dx/file',
			'Dir/Ex/file',
			'Dir/Fx/file',
			'Dir/Target/file',
		])))

	def test_2_4_recursive_links(self):
		"""
		Tests detection of recursive links.
		"""
		self.require_symlink()
		self.require_realpath()
		self.make_dirs([
			'Dir',
		])
		self.make_files([
			'Dir/file',
		])
		self.make_links([
			('Dir/Self', 'Dir'),
		])
		with self.assertRaises(RecursionError) as context:
			set(iter_tree_files(self.temp_dir))
		self.assertEqual(context.exception.first_path, 'Dir')
		self.assertEqual(context.exception.second_path, self.ospath('Dir/Self'))

	def test_2_5_recursive_circular_links(self):
		"""
		Tests detection of recursion through circular links.
		"""
		self.require_symlink()
		self.require_realpath()
		self.make_dirs([
			'A',
			'B',
			'C',
		])
		self.make_files([
			'A/d',
			'B/e',
			'C/f',
		])
		self.make_links([
			('A/Bx', 'B'),
			('B/Cx', 'C'),
			('C/Ax', 'A'),
		])
		with self.assertRaises(RecursionError) as context:
			set(iter_tree_files(self.temp_dir))
		self.assertIn(context.exception.first_path, ('A', 'B', 'C'))
		self.assertEqual(context.exception.second_path, {
			'A': self.ospath('A/Bx/Cx/Ax'),
			'B': self.ospath('B/Cx/Ax/Bx'),
			'C': self.ospath('C/Ax/Bx/Cx'),
		}[context.exception.first_path])

	def test_2_6_detect_broken_links(self):
		"""
		Tests that broken links are detected.
		"""
		def reraise(e):
			raise e

		self.require_symlink()
		self.make_links([
			('A', 'DOES_NOT_EXIST'),
		])
		with self.assertRaises(OSError) as context:
			set(iter_tree_files(self.temp_dir, on_error=reraise))
		self.assertEqual(context.exception.errno, errno.ENOENT)

	def test_2_7_ignore_broken_links(self):
		"""
		Tests that broken links are ignored.
		"""
		self.require_symlink()
		self.make_links([
			('A', 'DOES_NOT_EXIST'),
		])
		results = set(iter_tree_files(self.temp_dir))
		self.assertEqual(results, set())

	def test_2_8_no_follow_links(self):
		"""
		Tests to make sure directory links can be ignored.
		"""
		self.require_symlink()
		self.make_dirs([
			'Dir',
		])
		self.make_files([
			'A',
			'B',
			'Dir/C',
			'Dir/D',
		])
		self.make_links([
			('Ax', 'A'),
			('Bx', 'B'),
			('Dir/Cx', 'Dir/C'),
			('Dir/Dx', 'Dir/D'),
			('DirX', 'Dir'),
		])
		results = set(iter_tree_files(self.temp_dir, follow_links=False))
		self.assertEqual(results, set(map(self.ospath, [
			'A',
			'Ax',
			'B',
			'Bx',
			'Dir/C',
			'Dir/Cx',
			'Dir/D',
			'Dir/Dx',
			'DirX',
		])))

	def test_3_entries(self):
		"""
		Tests to make sure all files are found.
		"""
		self.make_dirs([
			'Empty',
			'Dir',
			'Dir/Inner',
		])
		self.make_files([
			'a',
			'b',
			'Dir/c',
			'Dir/d',
			'Dir/Inner/e',
			'Dir/Inner/f',
		])
		results = {entry.path for entry in iter_tree_entries(self.temp_dir)}
		self.assertEqual(results, set(map(self.ospath, [
			'a',
			'b',
			'Dir',
			'Dir/c',
			'Dir/d',
			'Dir/Inner',
			'Dir/Inner/e',
			'Dir/Inner/f',
			'Empty',
		])))

	@unittest.skipIf(sys.version_info < (3, 4), "pathlib entered stdlib in Python 3.4")
	def test_4_normalizing_pathlib_path(self):
		"""
		Tests passing pathlib.Path as argument.
		"""
		from pathlib import Path
		first_spec = normalize_file(Path('a.txt'))
		second_spec = normalize_file('a.txt')
		self.assertEqual(first_spec, second_spec)



"""
Naming checkers
"""
import re
from collections import defaultdict
from pathlib import Path

from robot.api import Token
from robot.parsing.model.statements import Arguments, KeywordCall
from robot.parsing.model.blocks import Keyword

from robocop.checkers import VisitorChecker
from robocop.rules import Rule, RuleParam, RuleSeverity
from robocop.utils import (
    ROBOT_VERSION,
    find_robot_vars,
    keyword_col,
    normalize_robot_name,
    normalize_robot_var_name,
    pattern_type,
    remove_robot_vars,
    token_col,
)

rules = {
    "0301": Rule(
        RuleParam(
            name="pattern",
            default=re.compile(r"[\.\?]"),
            converter=pattern_type,
            desc="pattern defining characters (not) allowed in a name",
        ),
        rule_id="0301",
        name="not-allowed-char-in-name",
        msg="Not allowed character '{{ character }}' found in {{ block_name }} name",
        severity=RuleSeverity.WARNING,
        docs="""
        Reports not allowed character found in Suite, Test Case or Keyword names. By default it's dot (`.`). You can 
        configure what characters are reported by calling::
        
             robocop --configure not-allowed-char-in-name:pattern:regex_pattern
             
        `regex_pattern` should define regex pattern for characters not allowed in names. For example `[@\[]` pattern 
        reports any occurrence of `@[` characters.
        """,
    ),
    "0302": Rule(
        RuleParam(
            name="convention",
            default="each_word_capitalized",
            converter=str,
            desc="possible values: 'each_word_capitalized' (default) or 'first_word_capitalized'",
        ),
        rule_id="0302",
        name="wrong-case-in-keyword-name",
        msg="Keyword name '{{ keyword_name }}' should use title case",
        severity=RuleSeverity.WARNING,
    ),
    "0303": Rule(
        rule_id="0303",
        name="keyword-name-is-reserved-word",
        msg="'{{ keyword_name }}' is a reserved keyword{{ error_msg }}",
        severity=RuleSeverity.ERROR,
        docs="""
        Do not use reserved names for keyword names. Following names are reserved:
        
          - IF
          - ELSE IF
          - ELSE
          - FOR
          - END
          - WHILE
          - CONTINUE
          - RETURN
          - TRY
          - EXCEPT

        """,
    ),
    "0305": Rule(
        rule_id="0305",
        name="underscore-in-keyword-name",
        msg="Underscores in keyword name '{{ keyword_name }}' can be replaced with spaces",
        severity=RuleSeverity.WARNING,
        docs="""
        Example::
        
            # bad
            keyword_with_underscores
            
            # good
            Keyword Without Underscores

        """,
    ),
    "0306": Rule(
        rule_id="0306",
        name="setting-name-not-in-title-case",
        msg="Setting name '{{ setting_name }}' should use title or upper case",
        severity=RuleSeverity.WARNING,
        docs="""
        Good::
        
             *** Settings ***
             Resource    file.resource
             
             *** Test Cases ***
             Test
                 [DOCUMENTATION]  Some documentation
                 Step
        
        Bad::
        
             *** Settings ***
             resource    file.resource
             
             *** Test Cases ***
             Test
                 [documentation]  Some documentation
                 Step
            
        """,
    ),
    "0307": Rule(
        rule_id="0307",
        name="section-name-invalid",
        msg="Section name should be in format '{{ section_title_case }}' or '{{ section_upper_case }}'",
        severity=RuleSeverity.WARNING,
        docs="""
        Good::
        
            *** SETTINGS ***
            *** Keywords ***
        
        Bad::
        
            *** keywords ***

        """,
    ),
    "0308": Rule(
        rule_id="0308",
        name="not-capitalized-test-case-title",
        msg="Test case '{{ test_name }}' title should start with capital letter",
        severity=RuleSeverity.WARNING,
    ),
    "0309": Rule(
        rule_id="0309",
        name="section-variable-not-uppercase",
        msg="Section variable '{{ variable_name }}' name should be uppercase",
        severity=RuleSeverity.WARNING,
    ),
    "0310": Rule(
        rule_id="0310",
        name="non-local-variables-should-be-uppercase",
        msg="Test, suite and global variables should be uppercase",
        severity=RuleSeverity.WARNING,
    ),
    "0311": Rule(
        rule_id="0311",
        name="else-not-upper-case",
        msg="ELSE and ELSE IF should be upper case",
        severity=RuleSeverity.ERROR,
    ),
    "0312": Rule(
        rule_id="0312",
        name="keyword-name-is-empty",
        msg="Keyword name should not be empty",
        severity=RuleSeverity.ERROR,
    ),
    "0313": Rule(
        rule_id="0313",
        name="test-case-name-is-empty",
        msg="Test case name should not be empty",
        severity=RuleSeverity.ERROR,
    ),
    "0314": Rule(
        rule_id="0314",
        name="empty-library-alias",
        msg="Library alias should not be empty",
        severity=RuleSeverity.ERROR,
        docs="""
        Use non-empty name when using library import with alias.
        
        Good::
           
            *** Settings ***
            Library  CustomLibrary  AS  AnotherName
        
        Bad::
        
             *** Settings ***
             Library  CustomLibrary  AS
        
        """,
    ),
    "0315": Rule(
        rule_id="0315",
        name="duplicated-library-alias",
        msg="Library alias should not be the same as original name",
        severity=RuleSeverity.WARNING,
        docs="""
        Example of rule violation::
        
             *** Settings ***
             Library  CustomLibrary  AS  CustomLibrary  # same as library name
             Library  CustomLibrary  AS  Custom Library  # same as library name (spaces are ignored)
        
        """,
    ),
    "0316": Rule(
        rule_id="0316",
        name="possible-variable-overwriting",
        msg="Variable '{{ variable_name }}' may overwrite similar variable inside '{{ block_name }}' {{ block_type }}. "
        "Note that variables are case-insensitive, and also spaces and underscores are ignored.",
        severity=RuleSeverity.INFO,
    ),
    "0317": Rule(
        rule_id="0317",
        name="hyphen-in-variable-name",
        msg="Use underscore in variable name '{{ variable_name }}' instead of hyphens to "
        "avoid treating them like minus sign",
        severity=RuleSeverity.INFO,
        docs="""
        Robot Framework supports evaluation of Python code inside ${ } brackets. For example::
        
             ${var2}  Set Variable  ${${var}-${var2}}
        
        That's why there is possibility that hyphen in name is not recognized as part of name but as minus sign.
        Better to use underscore (if it's intended)::
        
        ${var2}  Set Variable  ${ ${var}_${var2}}
        """,
    ),
    "0318": Rule(
        rule_id="0318",
        name="bdd-without-keyword-call",
        msg="BDD reserved keyword '{{ keyword_name }}' not followed by any keyword{{ error_msg }}",
        severity=RuleSeverity.WARNING,
        docs="""
        When using BDD reserved keywords (such as `GIVEN`, `WHEN`, `AND`, `BUT` or `THEN`) use them together with 
        name of the keyword to run. 
        
        Good::
        
            Given Setup Is Complete
            When User Log In
            Then User Should See Welcome Page
        
        Bad::
        
            Given
            When User Log In
            Then User Should See Welcome Page
        
        Since those words are used for BDD style it's also recommended not to use them within the keyword name.
        """,
    ),
    "0319": Rule(
        rule_id="0319",
        name="deprecated-statement",
        msg="'{{ statement_name }}' is deprecated since Robot Framework version "
        "{{ version }}, use '{{ alternative }}' instead",
        severity=RuleSeverity.WARNING,
        version=">=4.0",
    ),
}


class InvalidCharactersInNameChecker(VisitorChecker):
    """Checker for invalid characters in suite, test case or keyword name."""

    reports = ("not-allowed-char-in-name",)

    def visit_File(self, node):
        source = node.source if node.source else self.source
        if source:
            suite_name = Path(source).stem
            if "__init__" in suite_name:
                suite_name = Path(source).parent.name
            for char in suite_name:
                if self.param("not-allowed-char-in-name", "pattern").search(char):
                    self.report("not-allowed-char-in-name", character=char, block_name="suite", node=node)
        super().visit_File(node)

    def check_if_char_in_node_name(self, node, name_of_node, is_keyword=False):
        variables = find_robot_vars(node.name) if is_keyword else []
        index = 0
        while index < len(node.name):
            # skip variables
            if variables and variables[0][0] == index:
                start, stop = variables.pop(0)
                index += stop - start
            else:
                if self.param("not-allowed-char-in-name", "pattern").search(node.name[index]):
                    self.report(
                        "not-allowed-char-in-name",
                        character=node.name[index],
                        block_name=f"'{node.name}' {name_of_node}",
                        node=node,
                        col=node.col_offset + index + 1,
                    )
                index += 1

    def visit_TestCaseName(self, node):  # noqa
        self.check_if_char_in_node_name(node, "test case")

    def visit_KeywordName(self, node):  # noqa
        self.check_if_char_in_node_name(node, "keyword", is_keyword=True)


def reserved_error_msg(name, reserved_type):
    return (
        f". It must be in uppercase ({name.upper()}) when used as a marker with {reserved_type}. "
        f"Each marker should have minimum of 2 spaces as separator."
    )


def uppercase_error_msg(name):
    return f". It must be in uppercase ({name.upper()}) when used as a statement"


class KeywordNamingChecker(VisitorChecker):
    """Checker for keyword naming violations."""

    reports = (
        "wrong-case-in-keyword-name",
        "keyword-name-is-reserved-word",
        "underscore-in-keyword-name",
        "else-not-upper-case",
        "keyword-name-is-empty",
        "bdd-without-keyword-call",
    )
    reserved_words = {
        3: {
            "for": reserved_error_msg("for", "'FOR' loop"),
            "end": reserved_error_msg("end", "'FOR' loop"),
        },
        4: {
            "if": uppercase_error_msg("if"),
            "else if": uppercase_error_msg("else if"),
            "else": uppercase_error_msg("else"),
            "for": reserved_error_msg("for", "'FOR' loop"),
            "end": reserved_error_msg("end", "'FOR' or 'IF'"),
        },
        5: {
            "if": uppercase_error_msg("if"),
            "else if": uppercase_error_msg("else if"),
            "else": uppercase_error_msg("else"),
            "for": reserved_error_msg("for", "'FOR' loop"),
            "end": reserved_error_msg("end", "'FOR', 'IF' or 'TRY EXCEPT'"),
            "while": uppercase_error_msg("while"),
            "continue": uppercase_error_msg("continue"),
            "return": uppercase_error_msg("return"),
            "try": reserved_error_msg("try", "'TRY EXCEPT'"),
            "except": reserved_error_msg("except", "'TRY EXCEPT'"),
            "finally": reserved_error_msg("finally", "'TRY EXCEPT'"),
        },
    }
    else_statements = {"else", "else if"}
    bdd = {"given", "when", "and", "but", "then"}

    def __init__(self):
        self.letter_pattern = re.compile(r"\W|_", re.UNICODE)
        self.inside_if_block = False
        super().__init__()

    def visit_Setup(self, node):  # noqa
        self.check_keyword_naming(node.name, node)

    visit_TestTeardown = visit_SuiteTeardown = visit_Teardown = visit_TestSetup = visit_SuiteSetup = visit_Setup

    def visit_Keyword(self, node):  # noqa
        if not node.name:
            self.report("keyword-name-is-empty", node=node)
        else:
            self.check_keyword_naming(node.name, node)
        self.generic_visit(node)

    def visit_KeywordCall(self, node):  # noqa
        if self.inside_if_block and node.keyword and node.keyword.lower() in self.else_statements:
            self.report("else-not-upper-case", node=node, col=keyword_col(node))
        self.check_keyword_naming(node.keyword, node)

    def visit_If(self, node):  # noqa
        self.inside_if_block = True
        self.generic_visit(node)
        self.inside_if_block = False

    def check_keyword_naming(self, keyword_name, node):  # noqa
        if not keyword_name or keyword_name.lstrip().startswith("#"):
            return
        if keyword_name == r"/":  # old for loop, / are interpreted as keywords
            return
        if (
            isinstance(node, KeywordCall)
            and normalize_robot_name(keyword_name, remove_prefix="builtin.") == "runkeywordif"
        ):
            for token in node.data_tokens:
                if (token.value.lower() in self.else_statements) and not token.value.isupper():
                    self.report(
                        "keyword-name-is-reserved-word",
                        keyword_name=token.value,
                        error_msg=reserved_error_msg(token.value, "'Run Keyword If'"),
                        node=node,
                        col=token.col_offset + 1,
                    )
        elif self.check_if_keyword_is_reserved(keyword_name, node):
            return
        self.check_bdd_keywords(keyword_name, node)
        normalized = remove_robot_vars(keyword_name)
        normalized = normalized.split(".")[-1]  # remove any imports ie ExternalLib.SubLib.Log -> Log
        normalized = normalized.replace("'", "")  # replace ' apostrophes
        if "_" in normalized:
            self.report("underscore-in-keyword-name", keyword_name=keyword_name, node=node)
        words = self.letter_pattern.sub(" ", normalized).split(" ")
        if self.param("wrong-case-in-keyword-name", "convention") == "first_word_capitalized":
            words = words[:1]
        if any(word[0].islower() for word in words if word):
            self.report("wrong-case-in-keyword-name", keyword_name=keyword_name, node=node)

    def check_bdd_keywords(self, keyword_name, node):
        if keyword_name.lower() not in self.bdd or isinstance(node, Keyword):
            return
        arg = node.get_token(Token.ARGUMENT)
        suffix = f". Use one space between: '{keyword_name.title()} {arg.value}'" if arg else ""
        col = token_col(node, Token.NAME, Token.KEYWORD)
        self.report("bdd-without-keyword-call", keyword_name=keyword_name, error_msg=suffix, node=node, col=col)

    def check_if_keyword_is_reserved(self, keyword_name, node):
        # if there is typo in syntax, it is interpreted as keyword
        reserved = self.reserved_words[ROBOT_VERSION.major]
        if keyword_name.lower() not in reserved:
            return False
        if keyword_name.lower() in self.else_statements and self.inside_if_block:
            return False  # handled by else-not-upper-case
        error_msg = reserved[keyword_name.lower()]
        col = keyword_col(node) if isinstance(node, KeywordCall) else keyword_col(node.header)
        self.report(
            "keyword-name-is-reserved-word",
            keyword_name=keyword_name,
            error_msg=error_msg,
            node=node,
            col=col,
        )
        return True


class SettingsNamingChecker(VisitorChecker):
    """Checker for settings and sections naming violations."""

    reports = (
        "setting-name-not-in-title-case",
        "section-name-invalid",
        "empty-library-alias",
        "duplicated-library-alias",
    )

    def __init__(self):
        self.section_name_pattern = re.compile(r"\*\*\*\s.+\s\*\*\*")
        super().__init__()

    def visit_SectionHeader(self, node):  # noqa
        name = node.data_tokens[0].value
        if not self.section_name_pattern.match(name) or not (name.istitle() or name.isupper()):
            valid_name = f"*** {node.name.title()} ***"
            self.report(
                "section-name-invalid", section_title_case=valid_name, section_upper_case=valid_name.upper(), node=node
            )

    def visit_Setup(self, node):  # noqa
        self.check_setting_name(node.data_tokens[0].value, node)

    visit_SuiteSetup = (
        visit_TestSetup
    ) = (
        visit_Teardown
    ) = (
        visit_SuiteTeardown
    ) = (
        visit_TestTeardown
    ) = (
        visit_ForceTags
    ) = (
        visit_DefaultTags
    ) = (
        visit_ResourceImport
    ) = (
        visit_VariablesImport
    ) = visit_Documentation = visit_Tags = visit_Timeout = visit_Template = visit_Arguments = visit_Return = visit_Setup

    def visit_LibraryImport(self, node):  # noqa
        self.check_setting_name(node.data_tokens[0].value, node)
        with_name = node.get_token(Token.WITH_NAME)
        if with_name is None:
            for arg in node.get_tokens(Token.ARGUMENT):
                if arg.value and arg.value == "WITH NAME":
                    self.report("empty-library-alias", node=arg, col=arg.col_offset + 1)
        else:
            if node.alias.replace(" ", "") == node.name.replace(" ", ""):  # New Name == NewName
                name_token = node.get_tokens(Token.NAME)[-1]
                self.report(
                    "duplicated-library-alias",
                    node=name_token,
                    col=name_token.col_offset + 1,
                )

    def check_setting_name(self, name, node):
        if not (name.istitle() or name.isupper()):
            self.report("setting-name-not-in-title-case", setting_name=name, node=node)


class TestCaseNamingChecker(VisitorChecker):
    """Checker for test case naming violations."""

    reports = (
        "not-capitalized-test-case-title",
        "test-case-name-is-empty",
    )

    def visit_TestCase(self, node):  # noqa
        if not node.name:
            self.report("test-case-name-is-empty", node=node)
        elif not node.name[0].isupper():
            self.report("not-capitalized-test-case-title", test_name=node.name, node=node)


class VariableNamingChecker(VisitorChecker):
    """Checker for variable naming violations."""

    reports = (
        "section-variable-not-uppercase",
        "non-local-variables-should-be-uppercase",
        "hyphen-in-variable-name",
    )

    def __init__(self):
        self.set_variable_variants = {
            "settaskvariable",
            "settestvariable",
            "setsuitevariable",
            "setglobalvariable",
        }
        super().__init__()

    def visit_VariableSection(self, node):  # noqa
        for child in node.body:
            if not child.data_tokens:
                continue
            token = child.data_tokens[0]
            if token.type == Token.VARIABLE and token.value and not token.value.isupper():
                self.report(
                    "section-variable-not-uppercase",
                    variable_name=token.value,
                    lineno=token.lineno,
                    col=token.col_offset + 1,
                )

    def visit_KeywordCall(self, node):  # noqa
        for token in node.get_tokens(Token.ASSIGN):
            if "-" in token.value:
                self.report(
                    "hyphen-in-variable-name",
                    variable_name=token.value,
                    lineno=token.lineno,
                    col=token.col_offset + 1,
                )

        if not node.keyword:
            return
        if normalize_robot_name(node.keyword, remove_prefix="builtin.") in self.set_variable_variants:
            if len(node.data_tokens) < 2:
                return
            token = node.data_tokens[1]
            if token.type == Token.ARGUMENT and not token.value.isupper():
                self.report(
                    "non-local-variables-should-be-uppercase",
                    node=node,
                    col=token.col_offset + 1,
                )


class SimilarVariableChecker(VisitorChecker):
    """Checker for finding same variables with similar names."""

    reports = ("possible-variable-overwriting",)

    def __init__(self):
        self.variables = defaultdict(set)
        self.parent_name = ""
        self.parent_type = ""
        super().__init__()

    def visit_Keyword(self, node):  # noqa
        self.variables = defaultdict(set)
        self.parent_name = node.name
        self.parent_type = type(node).__name__
        self.visit_vars_and_find_similar(node)
        self.generic_visit(node)

    visit_TestCase = visit_Keyword

    def visit_KeywordCall(self, node):  # noqa
        tokens = node.get_tokens(Token.ASSIGN)
        self.find_similar_variables(tokens, node)

    def visit_For(self, node):  # noqa
        for var in node.variables:
            self.variables[normalize_robot_var_name(var)].add(var)
        self.generic_visit(node)

    visit_ForLoop = visit_For

    def visit_vars_and_find_similar(self, node):
        """
        Creates a dictionary `variables` with normalized variable name as a key
        and ads a list of all detected variations of this variable in the node as a value,
        then it checks if similar variable was found.
        """
        for child in node.body:
            # read arguments from Test Case or Keyword
            if isinstance(child, Arguments):
                for token in child.get_tokens(Token.ARGUMENT):
                    self.variables[normalize_robot_var_name(token.value)].add(token.value)

    def find_similar_variables(self, tokens, node):
        for token in tokens:
            normalized_token = normalize_robot_var_name(token.value)
            if normalized_token in self.variables and token.value not in self.variables[normalized_token]:
                self.report(
                    "possible-variable-overwriting",
                    variable_name=token.value,
                    block_name=self.parent_name,
                    block_type=self.parent_type,
                    node=node,
                    lineno=token.lineno,
                    col=token.col_offset,
                )
            self.variables[normalized_token].add(token.value)


class DeprecatedStatementChecker(VisitorChecker):
    """Checker for deprecated statements."""

    reports = ("deprecated-statement",)
    deprecated_keywords = {
        4: {"runkeywordunless": "IF"},
        5: {
            "runkeywordunless": "IF",
            "runkeywordif": "IF",
            "exitforloop": "BREAK",
            "exitforloopif": "IF and BREAK",
            "continueforloop": "CONTINUE",
            "continueforloopif": "IF and CONTINUE",
            "returnfromkeyword": "RETURN",
            "returnfromkeywordif": "IF and RETURN",
        },
    }

    def visit_KeywordCall(self, node):  # noqa
        self.check_if_keyword_is_deprecated(node.keyword, node)

    def visit_SuiteSetup(self, node):  # noqa
        self.check_if_keyword_is_deprecated(node.name, node)

    visit_TestSetup = visit_Setup = visit_SuiteTeardown = visit_TestTeardown = visit_Teardown = visit_SuiteSetup

    def visit_Return(self, node):  # noqa
        """For RETURN use visit_ReturnStatement - visit_Return will most likely visit RETURN in the future"""
        self.report(
            "deprecated-statement",
            statement_name="[Return]",
            alternative="RETURN",
            node=node,
            col=token_col(node, Token.RETURN),
            version="5.*",
        )

    def check_if_keyword_is_deprecated(self, keyword_name, node):
        normalized_keyword_name = normalize_robot_name(keyword_name, remove_prefix="builtin.")
        deprecated_statements = self.deprecated_keywords.get(ROBOT_VERSION.major, {})
        if normalized_keyword_name in deprecated_statements:
            alternative = deprecated_statements[normalized_keyword_name]
            col = token_col(node, Token.NAME, Token.KEYWORD)
            self.report(
                "deprecated-statement",
                statement_name=keyword_name,
                alternative=alternative,
                node=node,
                col=col,
                version=f"{ROBOT_VERSION.major}.*",
            )



# Original work Copyright 2018 Brainwy Software Ltda (Dual Licensed: LGPL / Apache 2.0)
# From https://github.com/fabioz/pyvmmonitor-core
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Defines a PluginManager (which doesn't really have plugins, only a registry of extension points
and implementations for such extension points).

To use, create the extension points you want (any class starting with 'EP') and register
implementations for those.

I.e.:
pm = PluginManager()
pm.register(EPFoo, FooImpl, keep_instance=True)
pm.register(EPBar, BarImpl, keep_instance=False)

Then, later, to use it, it's possible to ask for instances through the PluginManager API:

foo_instances = pm.get_implementations(EPFoo) # Each time this is called, new
                                              # foo_instances will be created
                                              
bar_instance = pm.get_instance(EPBar) # Each time this is called, the same bar_instance is returned.

Alternatively, it's possible to use a decorator to use a dependency injection pattern -- i.e.:
don't call me, I'll call you ;)

@inject(foo_instance=EPFoo, bar_instances=[EPBar])
def m1(foo_instance, bar_instances, pm):
    for bar in bar_instances:
        ...
    foo_instance.foo
    
"""
import functools
from pathlib import Path
from typing import TypeVar, Any, Dict, Type, Tuple, Optional, Union


def execfile(file, glob=None, loc=None):
    import tokenize

    with tokenize.open(file) as stream:
        contents = stream.read()

    exec(compile(contents + "\n", file, "exec"), glob, loc)


class NotInstanceError(RuntimeError):
    pass


class NotRegisteredError(RuntimeError):
    pass


class InstanceAlreadyRegisteredError(RuntimeError):
    pass


T = TypeVar("T")


class PluginManager(object):
    """
    This is a manager of plugins (which we refer to extension points and implementations).
    Mostly, we have a number of EPs (Extension Points) and implementations may be registered
    for those extension points.
    The PluginManager is able to provide implementations (through #get_implementations) which are
    not kept on being tracked and a special concept which keeps an instance alive for an extension
    (through #get_instance).
    """

    def __init__(self) -> None:
        self._ep_to_impls: Dict[Type, list] = {}
        self._ep_to_instance_impls: Dict[Tuple[Type, Optional[str]], list] = {}
        self._ep_to_context_to_instance: Dict[Type, dict] = {}
        self._name_to_ep: Dict[str, Type] = {}
        self.exited = False

    def load_plugins_from(self, directory: Path) -> int:
        found_files_with_plugins = 0
        filepath: Path
        for filepath in directory.iterdir():
            if filepath.is_file() and filepath.name.endswith(".py"):
                namespace: dict = {"__file__": str(filepath)}
                execfile(str(filepath), glob=namespace, loc=namespace)
                register_plugins = namespace.get("register_plugins")
                if register_plugins is not None:
                    found_files_with_plugins += 1
                    register_plugins(self)
        return found_files_with_plugins

    # This should be:
    # def get_implementations(self, ep: Type[T]) -> List[T]:
    # But isn't due to: https://github.com/python/mypy/issues/5374
    def get_implementations(self, ep: Union[Type, str]) -> list:
        assert not self.exited
        if isinstance(ep, str):
            ep = self._name_to_ep[ep]

        impls = self._ep_to_impls.get(ep, [])
        ret = []
        for class_, kwargs in impls:
            instance = class_(**kwargs)
            ret.append(instance)

        return ret

    def register(
        self,
        ep: Type,
        impl,
        kwargs: Optional[dict] = None,
        context: Optional[str] = None,
        keep_instance: bool = False,
    ):
        """
        :param ep:
        :param str impl:
            This is the full path to the class implementation.
        :param kwargs:
        :param context:
            If keep_instance is True, it's possible to register it for a given
            context.
        :param keep_instance:
            If True, it'll be only available through pm.get_instance and the
            instance will be kept for further calls.
            If False, it'll only be available through get_implementations.
        """
        if kwargs is None:
            kwargs = {}
        assert not self.exited
        if isinstance(ep, str):
            raise ValueError("Expected the actual EP class to be passed.")
        self._name_to_ep[ep.__name__] = ep

        if keep_instance:
            ep_to_instance_impls = self._ep_to_instance_impls
            impls = ep_to_instance_impls.get((ep, context))
            if impls is None:
                impls = ep_to_instance_impls[(ep, context)] = []
            else:
                raise InstanceAlreadyRegisteredError(
                    "Unable to override when instance is kept and an implementation "
                    "is already registered."
                )
        else:
            ep_to_impl = self._ep_to_impls
            impls = ep_to_impl.get(ep)
            if impls is None:
                impls = ep_to_impl[ep] = []

        impls.append((impl, kwargs))

    def unregister(
        self, ep: Type, context: Optional[str] = None, keep_instance: bool = False
    ):
        self._name_to_ep.pop(ep.__name__, None)
        if keep_instance:
            ep_to_instance_impls = self._ep_to_instance_impls
            ep_to_instance_impls.pop((ep, context), None)
        else:
            ep_to_impl = self._ep_to_impls
            ep_to_impl.pop(ep, None)

    def set_instance(self, ep: Type, instance, context=None) -> None:
        if isinstance(ep, str):
            raise ValueError("Expected the actual EP class to be passed.")
        self._name_to_ep[ep.__name__] = ep

        instances = self._ep_to_context_to_instance.get(ep)
        if instances is None:
            instances = self._ep_to_context_to_instance[ep] = {}
        instances[context] = instance

    def iter_existing_instances(self, ep: Union[Type, str]):
        if isinstance(ep, str):
            ep = self._name_to_ep[ep]

        return self._ep_to_context_to_instance[ep].values()

    def has_instance(self, ep: Union[Type, str], context=None):
        if isinstance(ep, str):
            ep_cls = self._name_to_ep.get(ep)
            if ep_cls is None:
                return False

        try:
            self.get_instance(ep, context)
            return True
        except NotRegisteredError:
            return False

    # This should be:
    # def get_instance(self, ep: Type[T], context=None) -> T:
    # But isn't due to: https://github.com/python/mypy/issues/5374
    def get_instance(self, ep: Union[Type, str], context: Optional[str] = None) -> Any:
        """
        Creates an instance in this plugin manager: Meaning that whenever a new EP is asked in
        the same context it'll receive the same instance created previously (and it'll be
        kept alive in the plugin manager).
        """
        if self.exited:
            raise AssertionError("PluginManager already exited")

        if isinstance(ep, str):
            ep = self._name_to_ep[ep]
        try:
            return self._ep_to_context_to_instance[ep][context]
        except KeyError:
            try:
                impls = self._ep_to_instance_impls[(ep, context)]
            except KeyError:
                found = False
                if context is not None:
                    found = True
                    try:
                        impls = self._ep_to_instance_impls[(ep, None)]
                    except KeyError:
                        found = False
                if not found:
                    if ep in self._ep_to_impls:
                        # Registered but not a kept instance.
                        raise NotInstanceError()
                    else:
                        # Not registered at all.
                        raise NotRegisteredError()
            assert len(impls) == 1
            class_, kwargs = impls[0]

            instances = self._ep_to_context_to_instance.get(ep)
            if instances is None:
                instances = self._ep_to_context_to_instance[ep] = {}

            ret = instances[context] = class_(**kwargs)
            return ret

    __getitem__ = get_instance

    def exit(self):
        self.exited = True
        self._ep_to_context_to_instance.clear()
        self._ep_to_impls.clear()


def inject(**inject_kwargs):
    def decorator(func):
        @functools.wraps(func)
        def inject_dec(*args, **kwargs):
            pm = kwargs.get("pm")
            if pm is None:
                raise AssertionError(
                    "pm argument with PluginManager not passed (required for @inject)."
                )

            for key, val in inject_kwargs.items():
                if key not in kwargs:
                    if val.__class__ is list:
                        kwargs[key] = pm.get_implementations(val[0])
                    else:
                        kwargs[key] = pm.get_instance(val)
            return func(*args, **kwargs)

        return inject_dec

    return decorator



"""
Parsing utils
"""
from robocop.utils.disablers import DisablersFinder
from robocop.utils.file_types import FileType, FileTypeChecker
from robocop.utils.misc import (
    ROBOT_VERSION,
    AssignmentTypeDetector,
    RecommendationFinder,
    find_robot_vars,
    is_suite_templated,
    issues_to_lsp_diagnostic,
    keyword_col,
    last_non_empty_line,
    modules_from_path,
    modules_from_paths,
    modules_in_current_dir,
    normalize_robot_name,
    normalize_robot_var_name,
    parse_assignment_sign_type,
    pattern_type,
    remove_robot_vars,
    token_col,
    get_section_name,
    get_errors,
)



from robot.api.parsing import ModelTransformer, Token, End, Comment, EmptyLine

try:
    from robot.api.parsing import InlineIfHeader
except ImportError:
    InlineIfHeader = None

from robotidy.decorators import check_start_end_line


class AddMissingEnd(ModelTransformer):
    """
    Add missing END token to FOR loops and IF statements.

    Following code:

        FOR    ${x}    IN    foo    bar
            Log    ${x}

    will be transformed to:

        FOR    ${x}    IN    foo    bar
            Log    ${x}
        END

    Supports global formatting params: ``--startline`` and ``--endline``.
    """

    def fix_block(self, node, expected_type):
        self.generic_visit(node)
        self.fix_header_name(node, expected_type)
        outside = []
        if not node.end:  # fix statement position only if END was missing
            node.body, outside = self.collect_inside_statements(node)
        self.fix_end(node)
        return (node, *outside)

    @check_start_end_line
    def visit_For(self, node):  # noqa
        return self.fix_block(node, Token.FOR)

    @check_start_end_line
    def visit_While(self, node):  # noqa
        return self.fix_block(node, Token.WHILE)

    @check_start_end_line
    def visit_Try(self, node):  # noqa
        self.generic_visit(node)
        if node.type != Token.TRY:
            return node
        self.fix_header_name(node, node.type)
        outside = []
        if not node.end:  # fix statement position only if END was missing
            node.body, outside = self.collect_inside_statements(node)
            try_branch = self.get_last_except(node)
            if try_branch:
                try_branch.body, outside_try = self.collect_inside_statements(try_branch)
                outside += outside_try

        self.fix_end(node)
        return (node, *outside)

    @check_start_end_line
    def visit_If(self, node):  # noqa
        self.generic_visit(node)
        if node.type != Token.IF:
            return node
        if InlineIfHeader and isinstance(node.header, InlineIfHeader):
            self.fix_header_name(node, "IF")
            return node
        self.fix_header_name(node, node.type)
        outside = []
        if not node.end:
            node.body, outside = self.collect_inside_statements(node)
            or_else = self.get_last_or_else(node)
            if or_else:
                or_else.body, outside_or_else = self.collect_inside_statements(or_else)
                outside += outside_or_else

        self.fix_end(node)
        return (node, *outside)

    def fix_end(self, node):
        """Fix END (missing END, End -> END, END position should be the same as FOR etc)."""
        if node.header.tokens[0].type == Token.SEPARATOR:
            indent = node.header.tokens[0]
        else:
            indent = Token(Token.SEPARATOR, self.formatting_config.separator)
        node.end = End([indent, Token(Token.END, Token.END), Token(Token.EOL)])

    @staticmethod
    def fix_header_name(node, header_name):
        node.header.data_tokens[0].value = header_name

    def collect_inside_statements(self, node):
        """Split statements from node for those that belong to it and outside nodes.

        In this example with missing END:
            FOR  ${i}  IN RANGE  10
                Keyword
            Other Keyword

        RF will store 'Other Keyword' inside FOR block even if it should be outside.
        """
        new_body = [[], []]
        is_outside = False
        starting_col = self.get_column(node)
        for child in node.body:
            if not isinstance(child, EmptyLine) and self.get_column(child) <= starting_col:
                is_outside = True
            new_body[is_outside].append(child)
        while new_body[0] and isinstance(new_body[0][-1], EmptyLine):
            new_body[1].insert(0, new_body[0].pop())
        return new_body

    @staticmethod
    def get_column(node):
        if hasattr(node, "header"):
            return node.header.data_tokens[0].col_offset
        if isinstance(node, Comment):
            token = node.get_token(Token.COMMENT)
            return token.col_offset
        if not node.data_tokens:
            return node.col_offset
        return node.data_tokens[0].col_offset

    @staticmethod
    def get_last_or_else(node):
        if not node.orelse:
            return None
        or_else = node.orelse
        while or_else.orelse:
            or_else = or_else.orelse
        return or_else

    @staticmethod
    def get_last_except(node):
        if not node.next:
            return None
        try_branch = node.next
        while try_branch.next:
            try_branch = try_branch.next
        return try_branch



# Some comment lines to move the function below
# Some comment lines to move the function below
# Some comment lines to move the function below
# Some comment lines to move the function below


def full_function():
    # Note that this function is not called, it's there just to make the mapping explicit.
    # The test case should stop at `a = 1` and then skip the `print('Skip this print')`.
    # map to Cell1, line 1
    a = 1  # map to Cell1, line 2
    print('Skip this print')  # map to Cell1, line 3
    print('TEST SUCEEDED')  # map to Cell1, line 4
    b = 2  # map to Cell1, line 5


if __name__ == '__main__':
    code = compile('''# line 1
a = 1  # line 2
print('Skip this print')  # line 3
print('TEST SUCEEDED')  # line 4
b = 2  # line 5
''', '<Cell1>', 'exec')
    exec(code)



# Original work Copyright 2018 Palantir Technologies, Inc. (MIT)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import uuid

from robocorp_ls_core.robotframework_log import get_logger

from .exceptions import (
    JsonRpcException,
    JsonRpcRequestCancelled,
    JsonRpcInternalError,
    JsonRpcMethodNotFound,
)


from concurrent import futures
from robocorp_ls_core.basic import implements
from robocorp_ls_core.protocols import IEndPoint, IFuture, IMonitor
from typing import Optional

log = get_logger(__name__)
JSONRPC_VERSION = "2.0"
CANCEL_METHOD = "$/cancelRequest"

FORCE_NON_THREADED_VERSION = False


def require_monitor(func):
    """
    To be used as a decorator.
    """
    func.__require_monitor__ = True
    return func


class Endpoint(object):

    SHOW_THREAD_DUMP_AFTER_TIMEOUT = 5

    def __init__(self, dispatcher, consumer, id_generator=lambda: str(uuid.uuid4())):
        """A JSON RPC endpoint for managing messages sent to/from the client.

        Args:
            dispatcher (dict): A dictionary of method name to handler function.
                The handler functions should return either the result or a callable that will be used to asynchronously
                compute the result.
            consumer (fn): A function that consumes JSON RPC message dicts and sends them to the client.
            id_generator (fn, optional): A function used to generate request IDs.
                Defaults to the string value of :func:`uuid.uuid4`.
            max_workers (int, optional): The number of workers in the asynchronous executor pool.
        """
        import os

        self._dispatcher = dispatcher
        self._consumer = consumer
        self._id_generator = id_generator

        self._client_request_futures = {}
        self._server_request_futures = {}

        # i.e.: 5 to 15 workers.
        max_workers = min(15, (os.cpu_count() or 1) + 4)
        self._executor_service = futures.ThreadPoolExecutor(max_workers=max_workers)

    def shutdown(self):
        self._executor_service.shutdown(wait=False)

    @implements(IEndPoint.notify)
    def notify(self, method: str, params=None):
        log.debug("Sending notification: %s %s", method, params)

        message = {"jsonrpc": JSONRPC_VERSION, "method": method}
        if params is not None:
            message["params"] = params

        self._consumer(message)

    @implements(IEndPoint.request)
    def request(self, method: str, params=None) -> IFuture:
        msg_id = self._id_generator()
        log.debug("Sending request with id %s: %s %s", msg_id, method, params)

        message = {"jsonrpc": JSONRPC_VERSION, "id": msg_id, "method": method}
        if params is not None:
            message["params"] = params

        request_future: IFuture = futures.Future()
        request_future.add_done_callback(self._cancel_callback(msg_id))

        self._server_request_futures[msg_id] = request_future
        self._consumer(message)

        return request_future

    def _cancel_callback(self, request_id):
        """Construct a cancellation callback for the given request ID."""

        def callback(future):
            if future.cancelled():
                raise AssertionError(
                    "Futures should not be cancelled. Use future.set_exception(JsonRpcRequestCancelled()) instead."
                )

            exc = future.exception()
            if isinstance(exc, JsonRpcRequestCancelled):
                self.notify(CANCEL_METHOD, {"id": request_id})

        return callback

    @implements(IEndPoint.consume)
    def consume(self, message):
        if "jsonrpc" not in message or message["jsonrpc"] != JSONRPC_VERSION:
            log.warning("Unknown message type %s", message)
            return

        if "id" not in message:
            log.debug("Handling notification from client %s", message)
            self._handle_notification(message["method"], message.get("params"))
        elif "method" not in message:
            log.debug("Handling response from client %s", message)
            self._handle_response(
                message["id"], message.get("result"), message.get("error")
            )
        else:
            try:
                log.debug("Handling request from client %s", message)
                self._handle_request(
                    message["id"], message["method"], message.get("params")
                )
            except JsonRpcException as e:
                log.exception("Failed to handle request %s", message["id"])
                self._consumer(
                    {
                        "jsonrpc": JSONRPC_VERSION,
                        "id": message["id"],
                        "error": e.to_dict(),
                    }
                )
            except Exception:  # pylint: disable=broad-except
                log.exception("Failed to handle request %s", message["id"])
                self._consumer(
                    {
                        "jsonrpc": JSONRPC_VERSION,
                        "id": message["id"],
                        "error": JsonRpcInternalError.of(sys.exc_info()).to_dict(),
                    }
                )

    def _handle_notification(self, method, params):
        """Handle a notification from the client."""
        if method == CANCEL_METHOD:
            self._handle_cancel_notification(params["id"])
            return

        try:
            handler = self._dispatcher[method]
        except KeyError:
            log.warning("Ignoring notification for unknown method %s", method)
            return

        try:
            handler_result = handler(params)
        except Exception:  # pylint: disable=broad-except
            log.exception("Failed to handle notification %s: %s", method, params)
            return

        if callable(handler_result):
            log.debug("Executing async notification handler %s", handler_result)
            notification_future = self._executor_service.submit(handler_result)
            notification_future.add_done_callback(
                self._notification_callback(method, params)
            )

    @staticmethod
    def _notification_callback(method, params):
        """Construct a notification callback for the given request ID."""

        def callback(future):
            try:
                future.result()
                log.debug(
                    "Successfully handled async notification %s %s", method, params
                )
            except Exception:  # pylint: disable=broad-except
                log.exception(
                    "Failed to handle async notification %s %s", method, params
                )

        return callback

    def _handle_cancel_notification(self, msg_id):
        """Handle a cancel notification from the client."""
        request_future = self._client_request_futures.pop(msg_id, None)

        if not request_future:
            log.warning(
                "Received cancel notification for unknown message id %s", msg_id
            )
            return

        # Will only work if the request hasn't started executing
        monitor: Optional[IMonitor] = getattr(request_future, "__monitor__", None)
        if monitor is not None:
            monitor.cancel()
        if request_future.cancel():
            log.debug("Cancelled request with id %s", msg_id)

    def _call_checking_time(self, func, **kwargs):
        from robocorp_ls_core import timeouts
        import threading
        import traceback

        timeout_tracker = timeouts.TimeoutTracker.get_singleton()
        curr_thread = threading.current_thread()
        timeout = self.SHOW_THREAD_DUMP_AFTER_TIMEOUT

        def on_timeout(*args, **kwargs):
            stack_trace = [
                "===============================================================================",
                f"Slow request (already took: {timeout}s). Showing thread dump.",
                "================================= Thread Dump =================================",
            ]

            for thread_id, stack in sys._current_frames().items():
                if thread_id != curr_thread.ident:
                    continue
                stack_trace.append(
                    "\n-------------------------------------------------------------------------------"
                )
                stack_trace.append(f" Thread {curr_thread}")
                stack_trace.append("")

                if "self" in stack.f_locals:
                    stack_trace.append(f"self: {stack.f_locals['self']}\n")

                for filename, lineno, name, line in traceback.extract_stack(stack):
                    stack_trace.append(
                        ' File "%s", line %d, in %s' % (filename, lineno, name)
                    )
                    if line:
                        stack_trace.append("   %s" % (line.strip()))
            stack_trace.append(
                "\n=============================== END Thread Dump ==============================="
            )
            log.critical("\n".join(stack_trace))

        if timeout > 0:
            with timeout_tracker.call_on_timeout(timeout, on_timeout, kwargs):
                return func(**kwargs)
        else:
            return func(**kwargs)

    def _handle_request(self, msg_id, method, params):
        """Handle a request from the client."""
        import time

        initial_time = time.time()
        try:
            handler = self._dispatcher[method]
        except KeyError:
            raise JsonRpcMethodNotFound.of(method)

        handler_result = handler(params)

        if callable(handler_result):
            kwargs = {}
            monitor = None
            if getattr(handler_result, "__require_monitor__", False):
                from robocorp_ls_core.jsonrpc.monitor import Monitor

                monitor = Monitor(f"Message: id: {msg_id}, method: {method}")
                kwargs["monitor"] = monitor
            log.debug("Executing async request handler %s", handler_result)

            if FORCE_NON_THREADED_VERSION:
                # I.e.: non-threaded version without breaking api.
                handler_result = handler_result(**kwargs)
                log.debug(
                    "Got result from synchronous request handler (in %.2fs): %s",
                    time.time() - initial_time,
                    handler_result,
                )
                self._consumer(
                    {"jsonrpc": JSONRPC_VERSION, "id": msg_id, "result": handler_result}
                )

            else:
                request_future = self._executor_service.submit(
                    self._call_checking_time, handler_result, **kwargs
                )
                if monitor is not None:
                    request_future.__monitor__ = monitor
                self._client_request_futures[msg_id] = request_future
                request_future.add_done_callback(self._request_callback(msg_id))
        elif isinstance(handler_result, futures.Future):
            log.debug("Request handler is already a future %s", handler_result)
            self._client_request_futures[msg_id] = handler_result
            handler_result.add_done_callback(self._request_callback(msg_id))
        else:
            log.debug(
                "Got result from synchronous request handler (in %.2fs): %s",
                time.time() - initial_time,
                handler_result,
            )
            self._consumer(
                {"jsonrpc": JSONRPC_VERSION, "id": msg_id, "result": handler_result}
            )

    def _request_callback(self, request_id):
        """Construct a request callback for the given request ID."""

        def callback(future):
            # Remove the future from the client requests map
            self._client_request_futures.pop(request_id, None)

            try:
                message = {"jsonrpc": JSONRPC_VERSION, "id": request_id}

                if future.cancelled():
                    raise JsonRpcRequestCancelled()

                message["result"] = future.result()
            except JsonRpcRequestCancelled as e:
                log.debug("Cancelled request: %s", request_id)
                message["error"] = e.to_dict()
            except JsonRpcException as e:
                log.exception("Failed to handle request %s", request_id)
                message["error"] = e.to_dict()
            except Exception:  # pylint: disable=broad-except
                log.exception("Failed to handle request %s", request_id)
                message["error"] = JsonRpcInternalError.of(sys.exc_info()).to_dict()

            self._consumer(message)

        return callback

    def _handle_response(self, msg_id, result=None, error=None):
        """Handle a response from the client."""
        request_future = self._server_request_futures.pop(msg_id, None)

        if not request_future:
            log.warning("Received response to unknown message id %s", msg_id)
            return

        if error is not None:
            log.debug("Received error response to message %s: %s", msg_id, error)
            request_future.set_exception(JsonRpcException.from_dict(error))
        else:
            log.debug("Received result for message %s: %s", msg_id, result)
            request_future.set_result(result)
